"""Regression tests for routing behavior.

These tests ensure that fixed bugs don't resurface, particularly:
1. Router should NOT respond when any agent is mentioned
2. Only mentioned agents should respond
"""

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import AsyncMock, MagicMock, patch

import nio
import pytest
from agno.models.ollama import Ollama

from mindroom.bot import AgentBot, _MessageContext
from mindroom.config.agent import AgentConfig
from mindroom.config.main import Config
from mindroom.config.models import ModelConfig, RouterConfig
from mindroom.matrix.users import AgentMatrixUser
from tests.conftest import TEST_PASSWORD

if TYPE_CHECKING:
    from pathlib import Path


def setup_test_bot(
    agent: AgentMatrixUser,
    storage_path: Path,
    room_id: str,
    enable_streaming: bool = False,
    config: Config | None = None,
) -> AgentBot:
    """Set up a test bot with all required mocks."""
    if config is None:
        config = Config.from_yaml()

    bot = AgentBot(agent, storage_path, rooms=[room_id], enable_streaming=enable_streaming, config=config)
    bot.client = AsyncMock()
    return bot


@pytest.fixture
def mock_research_agent() -> AgentMatrixUser:
    """Create a mock research agent user."""
    return AgentMatrixUser(
        agent_name="research",
        password=TEST_PASSWORD,
        display_name="MindRoomResearch",
        user_id="@mindroom_research:localhost",
    )


@pytest.fixture
def mock_news_agent() -> AgentMatrixUser:
    """Create a mock news agent user."""
    return AgentMatrixUser(
        agent_name="news",
        password=TEST_PASSWORD,
        display_name="MindRoomNews",
        user_id="@mindroom_news:localhost",
    )


class TestRoutingRegression:
    """Regression tests for routing behavior."""

    @pytest.mark.asyncio
    @patch("mindroom.bot.is_user_online")
    @patch("mindroom.bot.ai_response")
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_does_not_respond_when_agent_mentioned(
        self,
        mock_suggest_agent: AsyncMock,
        mock_ai_response: AsyncMock,
        mock_is_user_online: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Test that router doesn't activate when an agent is directly mentioned.

        Regression test for issue where both mentioned agent AND router-selected
        agent would respond to the same message.
        """
        test_room_id = "!research:localhost"

        # Mock user as online for stop button to show
        mock_is_user_online.return_value = True

        # Set up research bot (the one being mentioned)
        research_bot = setup_test_bot(mock_research_agent, tmp_path, test_room_id)

        # Set up news bot
        news_bot = setup_test_bot(mock_news_agent, tmp_path, test_room_id)

        # Mock AI responses
        mock_ai_response.return_value = "I can help with that research!"
        mock_suggest_agent.return_value = "news"  # Router would pick news

        # Mock successful room_send
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_123"
        research_bot.client.room_send.return_value = mock_send_response
        news_bot.client.room_send.return_value = mock_send_response

        # Create room with both agents
        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
        }

        # User mentions research agent specifically
        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@user:localhost"
        message_event.body = "@mindroom_research:localhost what can you do?"
        message_event.event_id = "$user_msg_123"
        message_event.source = {
            "content": {
                "body": "@mindroom_research:localhost what can you do?",
                "m.mentions": {"user_ids": ["@mindroom_research:localhost"]},
            },
        }

        # Process with research bot - SHOULD respond
        await research_bot._on_message(mock_room, message_event)
        assert research_bot.client.room_send.call_count == 3  # thinking + 🛑 + final
        assert mock_ai_response.call_count == 1

        # Process with news bot - should NOT respond and NOT use router
        await news_bot._on_message(mock_room, message_event)
        assert news_bot.client.room_send.call_count == 0
        # Router should NOT have been called
        assert mock_suggest_agent.call_count == 0

    @pytest.mark.asyncio
    @patch("mindroom.bot.ai_response")
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_activates_when_no_agent_mentioned(
        self,
        mock_suggest_agent: AsyncMock,
        mock_ai_response: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Test that router DOES activate when no agents are mentioned."""
        test_room_id = "!research:localhost"

        # Create test config with agents configured for the test room
        test_config = Config(
            agents={
                "research": AgentConfig(
                    display_name="MindRoomResearch",
                    rooms=[test_room_id],  # Configured for test room
                ),
                "news": AgentConfig(
                    display_name="MindRoomNews",
                    rooms=[test_room_id],  # Configured for test room
                ),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="test", id="test-model")},
            router=RouterConfig(model="default"),
        )

        # Create router agent
        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )

        # Set up router bot with test config
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id, config=test_config)

        # Set up research bot with test config
        research_bot = setup_test_bot(mock_research_agent, tmp_path, test_room_id, config=test_config)

        # Set up news bot with test config
        news_bot = setup_test_bot(mock_news_agent, tmp_path, test_room_id, config=test_config)

        # Mock AI responses
        mock_ai_response.return_value = "I can help with that!"
        mock_suggest_agent.return_value = "research"  # Router picks research

        # Mock successful room_send
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_123"
        router_bot.client.room_send.return_value = mock_send_response
        research_bot.client.room_send.return_value = mock_send_response
        news_bot.client.room_send.return_value = mock_send_response

        # Create room with all agents
        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            router_agent.user_id: MagicMock(),
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
        }

        # User message with NO mentions
        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@user:localhost"
        message_event.body = "What's the latest news?"
        message_event.event_id = "$user_msg_456"
        message_event.source = {
            "content": {
                "body": "What's the latest news?",
            },
        }

        # Process with router bot (should handle routing)
        await router_bot._on_message(mock_room, message_event)

        # Router SHOULD have been called
        mock_suggest_agent.assert_called_once()
        # Router bot should send the routing message
        assert router_bot.client.room_send.call_count == 1  # Router doesn't use stop button

        # Process with other bots - they should not do anything
        await research_bot._on_message(mock_room, message_event)
        await news_bot._on_message(mock_room, message_event)
        assert research_bot.client.room_send.call_count == 0
        assert news_bot.client.room_send.call_count == 0

    @pytest.mark.asyncio
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_filters_by_agent_reply_permissions(
        self,
        mock_suggest_agent: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Router should only route to agents that may reply to the current sender."""
        test_room_id = "!research:localhost"
        test_config = Config(
            agents={
                "research": AgentConfig(
                    display_name="MindRoomResearch",
                    rooms=[test_room_id],
                ),
                "news": AgentConfig(
                    display_name="MindRoomNews",
                    rooms=[test_room_id],
                ),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="test", id="test-model")},
            router=RouterConfig(model="default"),
            authorization={
                "default_room_access": True,
                "agent_reply_permissions": {
                    "research": ["@alice:localhost"],
                },
            },
        )

        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id, config=test_config)

        mock_suggest_agent.return_value = "news"
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_789"
        router_bot.client.room_send.return_value = mock_send_response

        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            router_agent.user_id: MagicMock(),
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
        }

        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@bob:localhost"
        message_event.body = "What's new today?"
        message_event.event_id = "$user_msg_789"
        message_event.source = {
            "content": {
                "body": "What's new today?",
            },
        }

        await router_bot._handle_ai_routing(
            mock_room,
            message_event,
            thread_history=[],
            requester_user_id="@bob:localhost",
        )

        mock_suggest_agent.assert_called_once()
        available_agents = mock_suggest_agent.call_args.args[1]
        assert [agent.agent_name(test_config) for agent in available_agents] == ["news"]

    @pytest.mark.asyncio
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_filters_by_agent_reply_permissions_with_multiple_allowed(
        self,
        mock_suggest_agent: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Router should pass all sender-allowed agents to routing when multiple are eligible."""
        test_room_id = "!research:localhost"
        test_config = Config(
            agents={
                "research": AgentConfig(
                    display_name="MindRoomResearch",
                    rooms=[test_room_id],
                ),
                "news": AgentConfig(
                    display_name="MindRoomNews",
                    rooms=[test_room_id],
                ),
                "facts": AgentConfig(
                    display_name="MindRoomFacts",
                    rooms=[test_room_id],
                ),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="test", id="test-model")},
            router=RouterConfig(model="default"),
            authorization={
                "default_room_access": True,
                "agent_reply_permissions": {
                    "research": ["@alice:localhost"],
                    "facts": ["@bob:localhost"],
                },
            },
        )

        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id, config=test_config)

        mock_suggest_agent.return_value = "news"
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_789_multi"
        router_bot.client.room_send.return_value = mock_send_response

        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            router_agent.user_id: MagicMock(),
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
            "@mindroom_facts:localhost": MagicMock(),
        }

        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@bob:localhost"
        message_event.body = "What's new today?"
        message_event.event_id = "$user_msg_789_multi"
        message_event.source = {
            "content": {
                "body": "What's new today?",
            },
        }

        await router_bot._on_message(mock_room, message_event)

        mock_suggest_agent.assert_called_once()
        available_agents = mock_suggest_agent.call_args.args[1]
        assert [agent.agent_name(test_config) for agent in available_agents] == ["facts", "news"]

    @pytest.mark.asyncio
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_reply_permissions_block_router_response(
        self,
        mock_suggest_agent: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Router should not respond when sender is disallowed for router replies."""
        test_room_id = "!research:localhost"
        test_config = Config(
            agents={
                "research": AgentConfig(
                    display_name="MindRoomResearch",
                    rooms=[test_room_id],
                ),
                "news": AgentConfig(
                    display_name="MindRoomNews",
                    rooms=[test_room_id],
                ),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="test", id="test-model")},
            router=RouterConfig(model="default"),
            authorization={
                "default_room_access": True,
                "agent_reply_permissions": {
                    "router": ["@alice:localhost"],
                    "research": ["*"],
                    "news": ["*"],
                },
            },
        )

        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id, config=test_config)

        mock_suggest_agent.return_value = "research"
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_791"
        router_bot.client.room_send.return_value = mock_send_response

        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            router_agent.user_id: MagicMock(),
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
        }

        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@bob:localhost"
        message_event.body = "What's new today?"
        message_event.event_id = "$user_msg_791"
        message_event.source = {
            "content": {
                "body": "What's new today?",
            },
        }

        await router_bot._on_message(mock_room, message_event)

        mock_suggest_agent.assert_not_called()
        router_bot.client.room_send.assert_not_called()

    @pytest.mark.asyncio
    @patch("mindroom.bot.suggest_agent_for_message")
    async def test_router_routes_when_thread_agents_are_disallowed_for_sender(
        self,
        mock_suggest_agent: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Router should still route in threads when prior agents cannot reply to this sender."""
        test_room_id = "!research:localhost"
        test_config = Config(
            agents={
                "research": AgentConfig(
                    display_name="MindRoomResearch",
                    rooms=[test_room_id],
                ),
                "news": AgentConfig(
                    display_name="MindRoomNews",
                    rooms=[test_room_id],
                ),
                "facts": AgentConfig(
                    display_name="MindRoomFacts",
                    rooms=[test_room_id],
                ),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="test", id="test-model")},
            router=RouterConfig(model="default"),
            authorization={
                "default_room_access": True,
                "agent_reply_permissions": {
                    "research": ["@alice:localhost"],
                    "news": ["@bob:localhost"],
                    "facts": ["@alice:localhost"],
                },
            },
        )

        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id, config=test_config)

        mock_suggest_agent.return_value = "research"
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_790"
        router_bot.client.room_send.return_value = mock_send_response

        mock_room = MagicMock()
        mock_room.room_id = test_room_id
        mock_room.users = {
            router_agent.user_id: MagicMock(),
            mock_research_agent.user_id: MagicMock(),
            mock_news_agent.user_id: MagicMock(),
            "@mindroom_facts:localhost": MagicMock(),
        }

        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = "@alice:localhost"
        message_event.body = "Can someone continue this thread?"
        message_event.event_id = "$user_msg_790"
        message_event.source = {
            "content": {
                "body": "Can someone continue this thread?",
            },
        }

        with patch.object(
            router_bot,
            "_extract_message_context",
            new=AsyncMock(
                return_value=_MessageContext(
                    am_i_mentioned=False,
                    is_thread=True,
                    thread_id="$thread_root",
                    thread_history=[
                        {
                            "sender": "@mindroom_news:localhost",
                            "body": "Latest update from news agent",
                        },
                    ],
                    mentioned_agents=[],
                    has_non_agent_mentions=False,
                ),
            ),
        ):
            await router_bot._on_message(mock_room, message_event)

        mock_suggest_agent.assert_called_once()
        available_agents = mock_suggest_agent.call_args.args[1]
        assert [agent.agent_name(test_config) for agent in available_agents] == ["facts", "research"]

    @pytest.mark.asyncio
    @patch("mindroom.teams.Team.arun")
    @patch("mindroom.bot.ai_response")
    @patch("mindroom.teams.get_model_instance")
    @patch("mindroom.config.main.Config.from_yaml")
    async def test_multiple_mentions_each_responds_once(
        self,
        mock_from_yaml: MagicMock,
        mock_get_model_instance: MagicMock,
        mock_ai_response: AsyncMock,
        mock_team_arun: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,
        tmp_path: Path,
    ) -> None:
        """Test that when multiple agents are mentioned, each responds exactly once."""
        # Create a mock config with proper models
        mock_config = Config(
            agents={
                "research": AgentConfig(display_name="ResearchAgent", rooms=["!research:localhost"]),
                "news": AgentConfig(display_name="NewsAgent", rooms=["!research:localhost"]),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="anthropic", id="claude-3-5-haiku-latest")},
        )
        mock_from_yaml.return_value = mock_config

        # Get the actual domain from config
        domain = mock_config.domain

        # Update mock agents to use the correct domain
        mock_research_agent.user_id = f"@mindroom_research:{domain}"
        mock_news_agent.user_id = f"@mindroom_news:{domain}"

        # Mock get_model_instance to return a mock model
        mock_model = Ollama(id="test-model")
        mock_get_model_instance.return_value = mock_model

        test_room_id = "!research:localhost"

        # Set up both bots with the same config
        research_bot = setup_test_bot(mock_research_agent, tmp_path, test_room_id, config=mock_config)
        news_bot = setup_test_bot(mock_news_agent, tmp_path, test_room_id, config=mock_config)

        # Create a shared orchestrator with both bots properly configured
        mock_orchestrator = MagicMock()
        mock_orchestrator.agent_bots = {
            "research": research_bot,
            "news": news_bot,
        }
        mock_orchestrator.current_config = mock_config
        mock_orchestrator.config = mock_config  # This is what teams.py uses

        # Set the orchestrator on both bots
        research_bot.orchestrator = mock_orchestrator
        news_bot.orchestrator = mock_orchestrator

        # Mock AI responses and team response
        mock_ai_response.side_effect = ["Research response!", "News response!"]
        mock_team_arun.return_value = "Team response"

        # Mock successful room_send
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$response_123"
        research_bot.client.room_send.return_value = mock_send_response
        news_bot.client.room_send.return_value = mock_send_response

        # Create room
        mock_room = MagicMock()
        mock_room.room_id = test_room_id

        # User mentions BOTH agents
        message_event = MagicMock(spec=nio.RoomMessageText)
        message_event.sender = f"@user:{domain}"
        message_event.body = f"@mindroom_research:{domain} and @mindroom_news:{domain}, what do you think?"
        message_event.event_id = "$user_msg_789"
        message_event.source = {
            "content": {
                "body": f"@mindroom_research:{domain} and @mindroom_news:{domain}, what do you think?",
                "m.mentions": {"user_ids": [f"@mindroom_research:{domain}", f"@mindroom_news:{domain}"]},
            },
        }

        # Process with both bots
        await research_bot._on_message(mock_room, message_event)
        await news_bot._on_message(mock_room, message_event)

        # With simplified team behavior: multiple mentions should form a team
        # The alphabetically first agent (news) handles team formation
        # The other agent (research) does not respond individually
        assert research_bot.client.room_send.call_count == 0  # No individual response
        assert news_bot.client.room_send.call_count == 2  # Team response (thinking + final)
        assert mock_team_arun.call_count == 1  # Team formed once

    @pytest.mark.asyncio
    @patch("mindroom.bot.is_user_online")
    @patch("mindroom.bot.ai_response")
    async def test_router_message_has_completion_marker(
        self,
        mock_ai_response: AsyncMock,
        mock_is_user_online: AsyncMock,
        mock_research_agent: AgentMatrixUser,
        mock_news_agent: AgentMatrixUser,  # noqa: ARG002
        tmp_path: Path,
    ) -> None:
        """Test that router messages trigger responses from mentioned agents.

        Regression test for potential issue where router mentions an agent but
        that agent ignores it.
        """
        test_room_id = "!research:localhost"

        # Mock user as online for stop button to show
        mock_is_user_online.return_value = True

        # Create router agent
        router_agent = AgentMatrixUser(
            agent_name="router",
            password=TEST_PASSWORD,
            display_name="RouterAgent",
            user_id="@mindroom_router:localhost",
        )

        # Set up router bot
        router_bot = setup_test_bot(router_agent, tmp_path, test_room_id)

        # Set up research bot (will be mentioned by router)
        research_bot = setup_test_bot(mock_research_agent, tmp_path, test_room_id)

        # Mock AI response
        mock_ai_response.return_value = "I can help with that research question!"

        # Mock successful room_send
        mock_send_response = MagicMock()
        mock_send_response.__class__ = nio.RoomSendResponse
        mock_send_response.event_id = "$router_msg"
        router_bot.client.room_send.return_value = mock_send_response
        research_bot.client.room_send.return_value = mock_send_response

        # Create room
        mock_room = MagicMock()
        mock_room.room_id = test_room_id

        # Simulate router message from router agent mentioning research
        # The router sends its messages
        router_message = MagicMock(spec=nio.RoomMessageText)
        router_message.sender = "@mindroom_router:localhost"
        router_message.body = "@research could you help with this?"
        router_message.event_id = "$router_msg"
        router_message.source = {
            "content": {
                "body": "@research could you help with this?",
                "m.mentions": {"user_ids": ["@mindroom_research:localhost"]},
            },
        }

        # Process router message with research bot
        await research_bot._on_message(mock_room, router_message)

        # Research bot SHOULD respond
        assert research_bot.client.room_send.call_count == 3  # thinking + 🛑 + final
        assert mock_ai_response.call_count == 1



"""Tests for ``scripts.check_module_privacy``."""

from __future__ import annotations

from typing import TYPE_CHECKING

from scripts.check_module_privacy import find_private_candidates, find_private_module_imports

if TYPE_CHECKING:
    from pathlib import Path


def _write(path: Path, content: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(content, encoding="utf-8")


def _symbols(project_root: Path) -> set[tuple[str, str]]:
    return {(symbol.module, symbol.name) for symbol in find_private_candidates(project_root)}


def _private_module_imports(project_root: Path) -> set[tuple[str, str]]:
    return {(issue.module, issue.imported_by) for issue in find_private_module_imports(project_root)}


def test_fastapi_route_functions_and_models_are_skipped(tmp_path: Path) -> None:
    """FastAPI route handlers and request/response models should not be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "api.py",
        """
from fastapi import APIRouter
from pydantic import BaseModel

router = APIRouter()

class RequestModel(BaseModel):
    value: int

@router.get("/health")
async def health() -> dict[str, str]:
    return {"status": "ok"}

def local_helper() -> int:
    return 1
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.api", "health") not in symbols
    assert ("pkg.api", "RequestModel") not in symbols
    assert ("pkg.api", "router") not in symbols
    assert ("pkg.api", "local_helper") in symbols


def test_fastapi_related_type_aliases_and_derived_models_are_skipped(tmp_path: Path) -> None:
    """Names only referenced from route signatures/decorators should be skipped."""
    _write(
        tmp_path / "src" / "pkg" / "api.py",
        """
from fastapi import APIRouter, Query
from pydantic import BaseModel
from typing import Annotated

router = APIRouter()

class BasePayload(BaseModel):
    name: str

class ExtendedPayload(BasePayload):
    age: int

RoomFilter = Annotated[str | None, Query(default=None)]

@router.get("/items", response_model=ExtendedPayload)
async def list_items(room_id: RoomFilter = None) -> ExtendedPayload:
    return ExtendedPayload(name="a", age=1)
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.api", "ExtendedPayload") not in symbols
    assert ("pkg.api", "RoomFilter") not in symbols
    assert ("pkg.api", "list_items") not in symbols


def test_typer_callbacks_are_skipped(tmp_path: Path) -> None:
    """Typer command callbacks should not be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "cli.py",
        """
import typer

app = typer.Typer()

@app.command()
def run() -> None:
    pass

def local_helper() -> int:
    return 1
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.cli", "app") not in symbols
    assert ("pkg.cli", "run") not in symbols
    assert ("pkg.cli", "local_helper") in symbols


def test_logger_variable_is_ignored(tmp_path: Path) -> None:
    """Module-level logger should be ignored by privacy checks."""
    _write(
        tmp_path / "src" / "pkg" / "mod.py",
        """
from logging import getLogger

logger = getLogger(__name__)

def local_helper() -> int:
    logger.info("x")
    return 1
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.mod", "logger") not in symbols
    assert ("pkg.mod", "local_helper") in symbols


def test_pyproject_console_entrypoint_is_skipped(tmp_path: Path) -> None:
    """Console-script entrypoints in pyproject.toml should not be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "cli.py",
        """
def main() -> int:
    return 0
""".strip()
        + "\n",
    )
    _write(
        tmp_path / "pyproject.toml",
        """
[project]
name = "example"
version = "0.1.0"
scripts.example = "pkg.cli:main"
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.cli", "main") not in symbols


def test_fastapi_include_router_dependency_symbols_are_skipped(tmp_path: Path) -> None:
    """FastAPI dependency callbacks used via include_router should not be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "api.py",
        """
from fastapi import APIRouter, Depends, FastAPI

app = FastAPI()
router = APIRouter()

async def verify_user() -> dict[str, str]:
    return {"id": "1"}

app.include_router(router, dependencies=[Depends(verify_user)])

def local_helper() -> int:
    return 1
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.api", "verify_user") not in symbols
    assert ("pkg.api", "local_helper") in symbols


def test_shell_uvicorn_entrypoint_is_skipped(tmp_path: Path) -> None:
    """Uvicorn entrypoints in shell scripts should not be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "server.py",
        """
def build_app() -> object:
    return object()

asgi = build_app()
""".strip()
        + "\n",
    )
    _write(
        tmp_path / "run-server.sh",
        """
#!/usr/bin/env bash
exec uvicorn pkg.server:asgi --host 0.0.0.0 --port 8000
""".strip()
        + "\n",
    )

    symbols = _symbols(tmp_path)
    assert ("pkg.server", "asgi") not in symbols


def test_private_module_imported_within_same_package_is_ignored(tmp_path: Path) -> None:
    """Private modules can be imported from within their own package subtree."""
    _write(
        tmp_path / "src" / "pkg" / "one" / "_internal.py",
        """
def helper() -> int:
    return 1
""".strip()
        + "\n",
    )
    _write(
        tmp_path / "src" / "pkg" / "one" / "public.py",
        """
from ._internal import helper

VALUE = helper()
""".strip()
        + "\n",
    )

    assert _private_module_imports(tmp_path) == set()


def test_private_module_imported_from_other_package_is_reported(tmp_path: Path) -> None:
    """Private modules imported outside their package subtree should be flagged."""
    _write(
        tmp_path / "src" / "pkg" / "one" / "_internal.py",
        """
VALUE = 1
""".strip()
        + "\n",
    )
    _write(
        tmp_path / "src" / "pkg" / "two" / "public.py",
        """
from pkg.one import _internal

VALUE = _internal.VALUE
""".strip()
        + "\n",
    )

    private_module_imports = _private_module_imports(tmp_path)
    assert ("pkg.one._internal", "pkg.two.public") in private_module_imports


def test_private_module_imported_only_by_tests_is_ignored(tmp_path: Path) -> None:
    """Only imports from src/ should count for private module detection."""
    _write(
        tmp_path / "src" / "pkg" / "_internal.py",
        """
def helper() -> int:
    return 1
""".strip()
        + "\n",
    )
    _write(
        tmp_path / "tests" / "test_private.py",
        """
from pkg._internal import helper

assert helper() == 1
""".strip()
        + "\n",
    )

    assert _private_module_imports(tmp_path) == set()



"""Matrix avatar management helpers."""

import io
import mimetypes
from pathlib import Path

import nio

from mindroom.logging_config import get_logger

logger = get_logger(__name__)


def _guess_avatar_content_type(avatar_path: Path) -> str:
    """Infer the upload MIME type for an avatar file."""
    guessed_type, _ = mimetypes.guess_type(avatar_path.name)
    if guessed_type and guessed_type.startswith("image/"):
        return guessed_type

    return {
        ".png": "image/png",
        ".jpg": "image/jpeg",
        ".jpeg": "image/jpeg",
        ".webp": "image/webp",
        ".gif": "image/gif",
        ".svg": "image/svg+xml",
        ".ico": "image/x-icon",
    }.get(avatar_path.suffix.lower(), "application/octet-stream")


async def _upload_avatar_file(
    client: nio.AsyncClient,
    avatar_path: Path,
) -> str | None:
    """Upload an avatar file to the Matrix server.

    Args:
        client: Authenticated Matrix client
        avatar_path: Path to the avatar image file

    Returns:
        The content URI if successful, None otherwise

    """
    if not avatar_path.exists():
        logger.warning(f"Avatar file not found: {avatar_path}")
        return None

    content_type = _guess_avatar_content_type(avatar_path)

    with avatar_path.open("rb") as f:
        avatar_data = f.read()

    file_size = len(avatar_data)

    def data_provider(_upload_monitor: object, _unused_data: object) -> io.BytesIO:
        return io.BytesIO(avatar_data)

    upload_result = await client.upload(
        data_provider=data_provider,
        content_type=content_type,
        filename=avatar_path.name,
        filesize=file_size,
    )

    # nio returns tuple (response, error)
    if isinstance(upload_result, tuple):
        upload_response, error = upload_result
        if error:
            logger.error(f"Upload error: {error}")
            return None
    else:
        upload_response = upload_result

    if not isinstance(upload_response, nio.UploadResponse):
        logger.error(f"Failed to upload avatar: {upload_response}")
        return None

    if not upload_response.content_uri:
        logger.error("Upload response missing content_uri")
        return None

    return str(upload_response.content_uri)


async def _set_avatar_from_file(
    client: nio.AsyncClient,
    avatar_path: Path,
) -> bool:
    """Set a user's avatar from a local file.

    Args:
        client: Authenticated Matrix client
        avatar_path: Path to the avatar image file

    Returns:
        True if successful, False otherwise

    """
    avatar_url = await _upload_avatar_file(client, avatar_path)
    if not avatar_url:
        return False

    response = await client.set_avatar(avatar_url)

    if isinstance(response, nio.ProfileSetAvatarResponse):
        logger.info(f"✅ Successfully set avatar for {client.user_id}")
        return True

    logger.error(f"Failed to set avatar for {client.user_id}: {response}")
    return False


async def check_and_set_avatar(
    client: nio.AsyncClient,
    avatar_path: Path,
    room_id: str | None = None,
) -> bool:
    """Check if user or room has an avatar and set it if they don't.

    Args:
        client: Authenticated Matrix client
        avatar_path: Path to the avatar image file
        room_id: Optional room ID for setting room avatar (if None, sets user avatar)

    Returns:
        True if avatar was already set or successfully set, False otherwise

    """
    if room_id:
        # Check room avatar
        response = await client.room_get_state_event(room_id, "m.room.avatar")
        if isinstance(response, nio.RoomGetStateEventResponse) and response.content and response.content.get("url"):
            logger.debug(f"Avatar already set for room {room_id}")
            return True
        # Set room avatar
        return await _set_room_avatar_from_file(client, room_id, avatar_path)
    # Check user avatar
    response = await client.get_profile(client.user_id)
    if isinstance(response, nio.ProfileGetResponse) and response.avatar_url:
        logger.debug(f"Avatar already set for {client.user_id}")
        return True
    # Set user avatar
    return await _set_avatar_from_file(client, avatar_path)


async def _set_room_avatar_from_file(
    client: nio.AsyncClient,
    room_id: str,
    avatar_path: Path,
) -> bool:
    """Set the avatar for a Matrix room from a file.

    Args:
        client: Authenticated Matrix client
        room_id: The room ID to set the avatar for
        avatar_path: Path to the avatar image file

    Returns:
        True if avatar was successfully set, False otherwise

    """
    avatar_url = await _upload_avatar_file(client, avatar_path)
    if not avatar_url:
        return False

    # Set room avatar using room state
    response = await client.room_put_state(
        room_id=room_id,
        event_type="m.room.avatar",
        content={"url": avatar_url},
    )

    if isinstance(response, nio.RoomPutStateResponse):
        logger.info(f"✅ Successfully set avatar for room {room_id}")
        return True

    logger.error(f"Failed to set avatar for room {room_id}: {response}")
    return False



"""Memory-related configuration models."""

from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, Field

from mindroom.config.models import EmbedderConfig

MemoryBackend = Literal["mem0", "file"]


class _MemoryEmbedderConfig(BaseModel):
    """Memory embedder configuration."""

    provider: str = Field(
        default="openai",
        description="Embedder provider (openai, ollama, huggingface, sentence_transformers, etc)",
    )
    config: EmbedderConfig = Field(default_factory=EmbedderConfig, description="Provider-specific config")


class _MemoryLLMConfig(BaseModel):
    """Memory LLM configuration."""

    provider: str = Field(default="ollama", description="LLM provider (ollama, openai, anthropic)")
    config: dict[str, Any] = Field(default_factory=dict, description="Provider-specific LLM config")


class _MemoryFileConfig(BaseModel):
    """File-backed memory configuration."""

    path: str | None = Field(
        default=None,
        description=(
            "Directory for file-backed memory. Relative paths resolve from the config "
            "directory. Defaults to <storage_path>/memory_files when omitted."
        ),
    )
    max_entrypoint_lines: int = Field(
        default=200,
        ge=1,
        description="Maximum number of lines to preload from MEMORY.md",
    )


class _MemoryAutoFlushBatchConfig(BaseModel):
    """Batching controls for background memory auto-flush."""

    max_sessions_per_cycle: int = Field(
        default=10,
        ge=1,
        description="Maximum sessions processed in one auto-flush loop iteration",
    )
    max_sessions_per_agent_per_cycle: int = Field(
        default=3,
        ge=1,
        description="Maximum sessions per agent processed in one auto-flush loop iteration",
    )


class _MemoryAutoFlushContextConfig(BaseModel):
    """Existing-memory context limits injected into extraction runs."""

    memory_snippets: int = Field(
        default=5,
        ge=0,
        description="Maximum number of MEMORY.md snippets included for extraction dedupe context",
    )
    snippet_max_chars: int = Field(
        default=400,
        ge=1,
        description="Maximum characters per included memory snippet",
    )


class _MemoryAutoFlushExtractorConfig(BaseModel):
    """Extraction limits for one background memory flush job."""

    no_reply_token: str = Field(
        default="NO_REPLY",
        description="Token indicating no durable memory should be written",
    )
    max_messages_per_flush: int = Field(
        default=20,
        ge=1,
        description="Maximum session chat messages considered by one extraction job",
    )
    max_chars_per_flush: int = Field(
        default=12000,
        ge=1,
        description="Maximum message characters considered by one extraction job",
    )
    max_extraction_seconds: int = Field(
        default=30,
        ge=1,
        description="Timeout for one extraction job before retrying in a later cycle",
    )
    include_memory_context: _MemoryAutoFlushContextConfig = Field(
        default_factory=_MemoryAutoFlushContextConfig,
        description="Bounds for existing memory context included during extraction",
    )


class MemoryAutoFlushConfig(BaseModel):
    """Background memory auto-flush configuration."""

    enabled: bool = Field(default=False, description="Enable background file-memory auto-flush worker")
    flush_interval_seconds: int = Field(
        default=1800,
        ge=5,
        description="Background auto-flush loop interval",
    )
    idle_seconds: int = Field(
        default=120,
        ge=0,
        description="Session idle time before dirty session becomes flush-eligible",
    )
    max_dirty_age_seconds: int = Field(
        default=600,
        ge=1,
        description="Force flush eligibility once a session remains dirty for this long",
    )
    stale_ttl_seconds: int = Field(
        default=86400,
        ge=60,
        description="Drop stale flush-state entries older than this TTL",
    )
    max_cross_session_reprioritize: int = Field(
        default=5,
        ge=0,
        description="Maximum same-agent dirty sessions reprioritized per incoming prompt",
    )
    retry_cooldown_seconds: int = Field(
        default=30,
        ge=1,
        description="Cooldown before retrying a failed extraction attempt",
    )
    max_retry_cooldown_seconds: int = Field(
        default=300,
        ge=1,
        description="Upper bound for retry cooldown backoff",
    )
    batch: _MemoryAutoFlushBatchConfig = Field(
        default_factory=_MemoryAutoFlushBatchConfig,
        description="Batch sizing controls for each auto-flush cycle",
    )
    extractor: _MemoryAutoFlushExtractorConfig = Field(
        default_factory=_MemoryAutoFlushExtractorConfig,
        description="Extraction-window and timeout controls for auto-flush",
    )


class MemoryConfig(BaseModel):
    """Memory system configuration."""

    backend: MemoryBackend = Field(
        default="mem0",
        description="Memory backend: 'mem0' (vector memory) or 'file' (markdown memory files)",
    )
    team_reads_member_memory: bool = Field(
        default=False,
        description=(
            "When true, team-context memory reads can access member agent memories in addition to the shared team scope"
        ),
    )
    embedder: _MemoryEmbedderConfig = Field(
        default_factory=_MemoryEmbedderConfig,
        description="Embedder configuration for memory",
    )
    llm: _MemoryLLMConfig | None = Field(default=None, description="LLM configuration for memory")
    file: _MemoryFileConfig = Field(default_factory=_MemoryFileConfig, description="File-backed memory configuration")
    auto_flush: MemoryAutoFlushConfig = Field(
        default_factory=MemoryAutoFlushConfig,
        description="Background auto-flush behavior for file-backed memory",
    )



"""Comprehensive HTTP API tests for webhook endpoints."""

from unittest.mock import MagicMock, Mock, patch

import pytest
import stripe
from fastapi.testclient import TestClient


class TestWebhookEndpoints:
    """Test webhook endpoints via HTTP API."""

    @pytest.fixture
    def client(self) -> TestClient:
        """Create test client."""
        from main import app  # noqa: PLC0415

        return TestClient(app)

    @pytest.fixture
    def mock_stripe_signature(self):
        """Mock Stripe signature verification."""
        with patch("backend.routes.webhooks.stripe.Webhook.construct_event") as mock:
            yield mock

    @pytest.fixture
    def mock_supabase(self):
        """Mock Supabase client."""
        with patch("backend.routes.webhooks.ensure_supabase") as mock:
            sb = MagicMock()
            mock.return_value = sb
            yield sb

    def _create_stripe_event(self, event_type: str, data: dict, event_id: str = "evt_test_123") -> Mock:
        """Create a mock Stripe event."""
        event = Mock()
        event.id = event_id
        event.type = event_type
        event.data.object = data
        return event

    def _create_subscription_data(
        self,
        subscription_id: str = "sub_test_123",
        customer_id: str = "cus_test_123",
        status: str = "active",
        tier: str = "starter",
        billing_cycle: str = "monthly",
        quantity: int = 1,
    ) -> dict:
        """Create test subscription data."""
        return {
            "id": subscription_id,
            "customer": customer_id,
            "status": status,
            "items": {
                "data": [
                    {
                        "price": {
                            "id": f"price_{tier}_{billing_cycle}",
                            "metadata": {
                                "plan": tier,
                                "billing_cycle": billing_cycle,
                            },
                        },
                        "quantity": quantity,
                    }
                ]
            },
            "current_period_start": 1700000000,
            "current_period_end": 1702678400,
            "trial_end": None,
        }

    def _create_invoice_data(
        self,
        invoice_id: str = "in_test_123",
        customer_id: str = "cus_test_123",
        subscription_id: str = "sub_test_123",
        amount_paid: int = 2900,  # in cents
        currency: str = "usd",
    ) -> dict:
        """Create test invoice data."""
        return {
            "id": invoice_id,
            "customer": customer_id,
            "subscription": subscription_id,
            "amount_paid": amount_paid,
            "currency": currency,
            "created": 1700000000,
            "billing_reason": "subscription_cycle",
        }

    def test_webhook_missing_signature(self, client: TestClient):
        """Test webhook endpoint without signature header."""
        response = client.post("/webhooks/stripe", json={})
        assert response.status_code == 400
        assert response.json()["detail"] == "Missing signature"

    def test_webhook_invalid_signature(self, client: TestClient, mock_stripe_signature: Mock):
        """Test webhook with invalid signature."""
        mock_stripe_signature.side_effect = stripe.error.SignatureVerificationError("Invalid signature", None)

        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "invalid_sig"},
        )
        assert response.status_code == 400
        assert response.json()["detail"] == "Invalid signature"

    def test_subscription_created_success(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test successful subscription creation webhook."""
        # Setup
        subscription_data = self._create_subscription_data()
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().select().eq().execute.return_value = Mock(data=[])
        mock_supabase.table().insert().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

        # Verify Supabase calls
        assert mock_supabase.table.call_count >= 3  # accounts, subscriptions check, insert

    def test_subscription_created_no_account(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test subscription creation with no matching account."""
        # Setup
        subscription_data = self._create_subscription_data()
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock no account found
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data=None)

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        assert response.json() == {
            "received": True,
            "error": "Failed to process subscription creation",
        }

    def test_subscription_updated_success(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test successful subscription update webhook."""
        # Setup
        subscription_data = self._create_subscription_data(tier="professional", quantity=5)
        event = self._create_stripe_event("customer.subscription.updated", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().update().eq().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_subscription_deleted_success(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test successful subscription deletion webhook."""
        # Setup
        subscription_data = {"id": "sub_test_123", "customer": "cus_test_123"}
        event = self._create_stripe_event("customer.subscription.deleted", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"account_id": "account_123"})
        mock_supabase.table().update().eq().eq().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_subscription_deleted_not_found(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test subscription deletion for non-existent subscription."""
        # Setup
        subscription_data = {"id": "sub_test_123", "customer": "cus_test_123"}
        event = self._create_stripe_event("customer.subscription.deleted", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock subscription not found
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data=None)

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        assert response.json() == {
            "received": True,
            "error": "Failed to process subscription deletion",
        }

    def test_payment_succeeded(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test successful payment webhook."""
        # Setup
        invoice_data = self._create_invoice_data()
        event = self._create_stripe_event("invoice.payment_succeeded", invoice_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().insert().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

        # Verify both payments and usage tables were updated
        insert_calls = mock_supabase.table().insert.call_count
        assert insert_calls >= 2  # payments + usage

    def test_payment_succeeded_no_subscription(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test payment webhook for one-time payment (no subscription)."""
        # Setup
        invoice_data = self._create_invoice_data(subscription_id=None)
        del invoice_data["subscription"]
        event = self._create_stripe_event("invoice.payment_succeeded", invoice_data)
        mock_stripe_signature.return_value = event

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify - should succeed but not process
        assert response.status_code == 200
        assert response.json() == {
            "received": True,
            "error": "Failed to process payment",
        }

    def test_payment_failed(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test failed payment webhook."""
        # Setup
        invoice_data = self._create_invoice_data()
        event = self._create_stripe_event("invoice.payment_failed", invoice_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"account_id": "account_123"})
        mock_supabase.table().update().eq().eq().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_payment_failed_no_subscription(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test failed payment webhook with no subscription found."""
        # Setup
        invoice_data = self._create_invoice_data()
        event = self._create_stripe_event("invoice.payment_failed", invoice_data)
        mock_stripe_signature.return_value = event

        # Mock no subscription found
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data=None)

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        assert response.json() == {
            "received": True,
            "error": "Failed to process payment failure",
        }

    def test_trial_will_end(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test trial ending webhook."""
        # Setup
        subscription_data = self._create_subscription_data()
        subscription_data["trial_end"] = 1700086400  # Tomorrow
        subscription_data["id"] = "sub_test_123"
        subscription_data["customer"] = "cus_test_123"
        event = self._create_stripe_event("customer.subscription.trial_will_end", subscription_data)
        # Set data.object as a dict-like object that the handler expects
        event.data.object = subscription_data
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_unhandled_event_type(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test unhandled webhook event type."""
        # Setup
        data_obj = {"id": "obj_123", "customer": "cus_test_123"}
        event = self._create_stripe_event("some.unhandled.event", data_obj)
        # Set data.object as the dict
        event.data.object = data_obj
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_webhook_processing_exception(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test webhook with processing exception."""
        # Setup
        subscription_data = self._create_subscription_data()
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase to raise exception
        mock_supabase.table().select().eq().single().execute.side_effect = Exception("Database error")

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        assert response.json() == {"received": True, "error": "Database error"}

    def test_webhook_event_recording_failure(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test webhook when event recording fails."""
        # Setup
        subscription_data = self._create_subscription_data()
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock successful processing but failed recording
        accounts_table = MagicMock()
        accounts_table.select().eq().single().execute.return_value = Mock(data={"id": "account_123"})

        subscriptions_table = MagicMock()
        subscriptions_table.select().eq().execute.return_value = Mock(data=[])
        subscriptions_table.insert().execute.return_value = Mock()

        webhook_table = MagicMock()
        webhook_table.insert().execute.side_effect = Exception("Recording failed")

        def table_side_effect(name):
            if name == "accounts":
                return accounts_table
            elif name == "subscriptions":
                return subscriptions_table
            elif name == "webhook_events":
                return webhook_table
            else:
                return MagicMock()

        mock_supabase.table.side_effect = table_side_effect

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify - should still succeed despite recording failure
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None  # No error despite recording failure

    def test_subscription_with_trial(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test subscription creation with trial period."""
        # Setup
        subscription_data = self._create_subscription_data()
        subscription_data["trial_end"] = 1702678400  # Future timestamp
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().select().eq().execute.return_value = Mock(data=[])
        mock_supabase.table().insert().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_professional_plan_scaling(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test professional plan with multiple users scales limits correctly."""
        # Setup
        subscription_data = self._create_subscription_data(tier="professional", quantity=10)
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().select().eq().execute.return_value = Mock(data=[])

        # Capture the insert call to verify scaled limits
        insert_data = None

        def capture_insert(data):
            nonlocal insert_data
            insert_data = data
            return Mock(execute=Mock(return_value=Mock()))

        mock_supabase.table().insert = capture_insert

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

        # Professional plan should scale by quantity
        # Base limits would be multiplied by 10
        assert insert_data is not None

    def test_subscription_update_with_cancellation(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test subscription update with cancellation timestamp."""
        # Setup
        subscription_data = self._create_subscription_data(status="canceled")
        subscription_data["canceled_at"] = 1700086400
        event = self._create_stripe_event("customer.subscription.updated", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().update().eq().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_rate_limiting(self, client: TestClient, mock_stripe_signature: Mock, mock_supabase: MagicMock):
        """Test webhook rate limiting."""
        # Setup a valid event
        event = self._create_stripe_event("customer.subscription.created", self._create_subscription_data())
        mock_stripe_signature.return_value = event

        # Mock Supabase responses for all requests
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().select().eq().execute.return_value = Mock(data=[])
        mock_supabase.table().insert().execute.return_value = Mock()

        # Make many requests quickly to trigger rate limit
        # The limit is 20/minute, so 21 requests should trigger it
        responses = []
        for _ in range(25):
            response = client.post(
                "/webhooks/stripe",
                content=b"test body",
                headers={"Stripe-Signature": "valid_sig"},
            )
            responses.append(response.status_code)

        # At least one should be rate limited (429)
        assert 429 in responses

    def test_legacy_tier_metadata(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test handling of legacy 'tier' metadata field."""
        # Setup with legacy metadata format
        subscription_data = self._create_subscription_data()
        # Use 'tier' instead of 'plan' in metadata
        subscription_data["items"]["data"][0]["price"]["metadata"] = {
            "tier": "starter",  # Legacy field name
            "billing_cycle": "monthly",
        }
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})
        mock_supabase.table().select().eq().execute.return_value = Mock(data=[])
        mock_supabase.table().insert().execute.return_value = Mock()

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify - should handle legacy format successfully
        assert response.status_code == 200
        data = response.json()
        assert data["received"] is True
        assert data["error"] is None

    def test_missing_price_metadata(
        self,
        client: TestClient,
        mock_stripe_signature: Mock,
        mock_supabase: MagicMock,
    ):
        """Test handling of missing price metadata."""
        # Setup with no metadata
        subscription_data = self._create_subscription_data()
        subscription_data["items"]["data"][0]["price"]["metadata"] = {}
        event = self._create_stripe_event("customer.subscription.created", subscription_data)
        mock_stripe_signature.return_value = event

        # Mock Supabase responses
        mock_supabase.table().select().eq().single().execute.return_value = Mock(data={"id": "account_123"})

        # Make request
        response = client.post(
            "/webhooks/stripe",
            content=b"test body",
            headers={"Stripe-Signature": "valid_sig"},
        )

        # Verify - should fail gracefully
        assert response.status_code == 200
        # Should have an error since tier couldn't be determined
        result = response.json()
        assert result["received"] is True
        assert "error" in result or "Unable to determine tier" in str(result)



"""Test threading behavior to reproduce and fix the threading error.

This test verifies that:
1. Agents always respond in threads (never in main room)
2. Commands that are replies don't cause threading errors
3. The bot handles various message relation scenarios correctly
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock, patch

import nio
import pytest
import pytest_asyncio

from mindroom.bot import AgentBot
from mindroom.config.agent import AgentConfig
from mindroom.config.main import Config
from mindroom.config.models import ModelConfig, RouterConfig
from mindroom.matrix.reply_chain import _merge_thread_and_chain_history
from mindroom.matrix.users import AgentMatrixUser
from tests.conftest import TEST_PASSWORD

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from pathlib import Path


class TestThreadingBehavior:
    """Test that agents correctly handle threading in various scenarios."""

    @pytest_asyncio.fixture
    async def bot(self, tmp_path: Path) -> AsyncGenerator[AgentBot, None]:
        """Create an AgentBot for testing."""
        agent_user = AgentMatrixUser(
            user_id="@mindroom_general:localhost",
            password=TEST_PASSWORD,
            display_name="GeneralAgent",
            agent_name="general",
        )

        config = Config(
            agents={"general": AgentConfig(display_name="GeneralAgent", rooms=["!test:localhost"])},
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="ollama", id="test-model")},
            router=RouterConfig(model="default"),
        )

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            rooms=["!test:localhost"],
            enable_streaming=False,  # Disable streaming for simpler testing
            config=config,
        )

        # Mock the orchestrator
        mock_orchestrator = MagicMock()
        mock_orchestrator.current_config = config
        bot.orchestrator = mock_orchestrator

        # Create a mock client
        bot.client = AsyncMock(spec=nio.AsyncClient)
        bot.client.rooms = {}
        bot.client.user_id = "@mindroom_general:localhost"
        bot.client.homeserver = "http://localhost:8008"

        # Initialize components that depend on client
        bot.response_tracker = MagicMock()
        bot.response_tracker.has_responded.return_value = False

        # Mock the agent to return a response
        mock_agent = MagicMock()
        mock_response = MagicMock()
        mock_response.content = "I can help you with that!"

        # Make the agent's arun method return the response
        async def mock_arun(*_args: object, **_kwargs: object) -> MagicMock:
            return mock_response

        mock_agent.arun = mock_arun

        # Mock create_agent to return our mock agent
        with patch("mindroom.bot.create_agent", return_value=mock_agent):
            yield bot

        # No cleanup needed since we're using mocks

    @pytest.mark.asyncio
    async def test_agent_creates_thread_when_mentioned_in_main_room(self, bot: AgentBot) -> None:
        """Test that agents create threads when mentioned in main room messages."""
        room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
        room.name = "Test Room"

        # Create a main room message that mentions the agent
        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "@mindroom_general Can you help me?",
                    "msgtype": "m.text",
                    "m.mentions": {"user_ids": ["@mindroom_general:localhost"]},
                },
                "event_id": "$main_msg:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        # The bot should send a response
        bot.client.room_send = AsyncMock(
            return_value=nio.RoomSendResponse.from_dict({"event_id": "$response:localhost"}, room_id="!test:localhost"),
        )

        # Mock thread history fetch (returns empty for new thread)
        bot.client.room_messages = AsyncMock(
            return_value=nio.RoomMessagesResponse.from_dict(
                {"chunk": [], "start": "s1", "end": "e1"},
                room_id="!test:localhost",
            ),
        )

        # Initialize the bot (to set up components it needs)
        bot.response_tracker.has_responded.return_value = False

        # Mock interactive.handle_text_response to return None (not an interactive response)
        # Mock _generate_response to capture the call and send a test response
        with (
            patch("mindroom.bot.interactive.handle_text_response", AsyncMock(return_value=None)),
            patch.object(bot, "_generate_response") as mock_generate,
        ):
            # Process the message
            await bot._on_message(room, event)

            # Check that _generate_response was called
            mock_generate.assert_called_once()

            # Now simulate the response being sent
            await bot._send_response(room.room_id, event.event_id, "I can help you with that!", None)

        # Verify the bot sent a response
        bot.client.room_send.assert_called_once()

        # Check the content of the response
        call_args = bot.client.room_send.call_args
        content = call_args.kwargs["content"]

        # The response should create a thread from the original message
        assert "m.relates_to" in content
        assert content["m.relates_to"]["rel_type"] == "m.thread"
        assert content["m.relates_to"]["event_id"] == "$main_msg:localhost"
        assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$main_msg:localhost"

    @pytest.mark.asyncio
    async def test_agent_responds_in_existing_thread(self, bot: AgentBot) -> None:
        """Test that agents respond correctly in existing threads."""
        room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
        room.name = "Test Room"

        # Create a message in a thread
        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "@mindroom_general What about this?",
                    "msgtype": "m.text",
                    "m.mentions": {"user_ids": ["@mindroom_general:localhost"]},
                    "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                },
                "event_id": "$thread_msg:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        # Mock the bot's response
        bot.client.room_send = AsyncMock(
            return_value=nio.RoomSendResponse.from_dict({"event_id": "$response:localhost"}, room_id="!test:localhost"),
        )

        # Mock thread history
        bot.client.room_messages = AsyncMock(
            return_value=nio.RoomMessagesResponse.from_dict(
                {"chunk": [], "start": "s1", "end": "e1"},
                room_id="!test:localhost",
            ),
        )

        # Initialize response tracking
        bot.response_tracker.has_responded.return_value = False

        # Mock interactive.handle_text_response and make AI fast
        with (
            patch("mindroom.bot.interactive.handle_text_response", AsyncMock(return_value=None)),
            patch("mindroom.bot.ai_response", AsyncMock(return_value="OK")),
            patch("mindroom.bot.get_latest_thread_event_id_if_needed", AsyncMock(return_value="latest_thread_event")),
        ):
            # Process the message
            await bot._on_message(room, event)

        # Verify the bot sent messages (thinking + final)
        assert bot.client.room_send.call_count == 2

        # Check the initial message (first call)
        first_call = bot.client.room_send.call_args_list[0]
        initial_content = first_call.kwargs["content"]
        assert "m.relates_to" in initial_content
        assert initial_content["m.relates_to"]["rel_type"] == "m.thread"
        assert initial_content["m.relates_to"]["event_id"] == "$thread_root:localhost"
        assert initial_content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$thread_msg:localhost"

    @pytest.mark.asyncio
    async def test_extract_context_maps_plain_reply_to_existing_thread(self, bot: AgentBot) -> None:
        """Plain replies to thread messages should resolve to the original thread root."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Follow-up from a non-thread client",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$thread_msg:localhost"}},
                },
                "event_id": "$reply_plain:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567891,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            return_value=nio.RoomGetEventResponse.from_dict(
                {
                    "content": {
                        "body": "Agent answer in thread",
                        "msgtype": "m.text",
                        "m.relates_to": {
                            "rel_type": "m.thread",
                            "event_id": "$thread_root:localhost",
                        },
                    },
                    "event_id": "$thread_msg:localhost",
                    "sender": "@mindroom_general:localhost",
                    "origin_server_ts": 1234567890,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            ),
        )

        expected_history = [
            {"event_id": "$thread_root:localhost", "body": "Root"},
            {"event_id": "$thread_msg:localhost", "body": "Agent answer in thread"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=expected_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$thread_root:localhost"
        assert context.thread_history == expected_history
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$thread_root:localhost")

    @pytest.mark.asyncio
    async def test_extract_context_maps_plain_reply_to_thread_root_with_existing_replies(self, bot: AgentBot) -> None:
        """Plain replies to a thread root should load full thread history, not just the root event."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Follow-up from a non-thread client",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$thread_root:localhost"}},
                },
                "event_id": "$reply_plain_root:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567892,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            return_value=nio.RoomGetEventResponse.from_dict(
                {
                    "content": {
                        "body": "Original root message",
                        "msgtype": "m.text",
                    },
                    "event_id": "$thread_root:localhost",
                    "sender": "@user:localhost",
                    "origin_server_ts": 1234567889,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            ),
        )

        expected_history = [
            {"event_id": "$thread_root:localhost", "body": "Original root message"},
            {"event_id": "$thread_msg:localhost", "body": "Agent answer in thread"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=expected_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$thread_root:localhost"
        assert context.thread_history == expected_history
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$thread_root:localhost")

    @pytest.mark.asyncio
    async def test_extract_context_edit_uses_thread_from_new_content(self, bot: AgentBot) -> None:
        """Edit events should resolve thread context from m.new_content thread relation."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "* updated",
                    "msgtype": "m.text",
                    "m.new_content": {
                        "body": "updated",
                        "msgtype": "m.text",
                        "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                    },
                    "m.relates_to": {"rel_type": "m.replace", "event_id": "$thread_msg:localhost"},
                },
                "event_id": "$edit_event:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567894,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        expected_history = [
            {"event_id": "$thread_root:localhost", "body": "Root"},
            {"event_id": "$thread_msg:localhost", "body": "Original"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=expected_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$thread_root:localhost"
        assert context.thread_history == expected_history
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$thread_root:localhost")

    @pytest.mark.asyncio
    async def test_extract_context_edit_resolves_thread_from_original_event(self, bot: AgentBot) -> None:
        """Edits without nested thread metadata should still resolve to the edited message thread."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "* updated",
                    "msgtype": "m.text",
                    "m.new_content": {
                        "body": "updated",
                        "msgtype": "m.text",
                    },
                    "m.relates_to": {"rel_type": "m.replace", "event_id": "$thread_msg:localhost"},
                },
                "event_id": "$edit_event:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567895,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            return_value=nio.RoomGetEventResponse.from_dict(
                {
                    "content": {
                        "body": "Thread message",
                        "msgtype": "m.text",
                        "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                    },
                    "event_id": "$thread_msg:localhost",
                    "sender": "@mindroom_general:localhost",
                    "origin_server_ts": 1234567893,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            ),
        )

        expected_history = [
            {"event_id": "$thread_root:localhost", "body": "Root"},
            {"event_id": "$thread_msg:localhost", "body": "Thread message"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=expected_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$thread_root:localhost"
        assert context.thread_history == expected_history
        bot.client.room_get_event.assert_awaited_once_with(room.room_id, "$thread_msg:localhost")
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$thread_root:localhost")

    @pytest.mark.asyncio
    async def test_extract_context_edit_keeps_reply_chain_context_without_threads(self, bot: AgentBot) -> None:
        """Reply-chain edits without thread metadata should keep linear context."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "* updated third message",
                    "msgtype": "m.text",
                    "m.new_content": {
                        "body": "updated third message",
                        "msgtype": "m.text",
                    },
                    "m.relates_to": {"rel_type": "m.replace", "event_id": "$msg3:localhost"},
                },
                "event_id": "$edit_msg3:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567896,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            side_effect=[
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Third message",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$msg2:localhost"}},
                        },
                        "event_id": "$msg3:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567893,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Second message",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$msg1:localhost"}},
                        },
                        "event_id": "$msg2:localhost",
                        "sender": "@mindroom_general:localhost",
                        "origin_server_ts": 1234567892,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "First message",
                            "msgtype": "m.text",
                        },
                        "event_id": "$msg1:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567891,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            ],
        )

        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$msg1:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == [
            "$msg1:localhost",
            "$msg2:localhost",
            "$msg3:localhost",
        ]
        assert bot.client.room_get_event.await_count == 3
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_builds_reply_chain_history_without_threads(self, bot: AgentBot) -> None:
        """Reply-only chains should still keep linear conversation context."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Third message in a reply-only chain",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$msg2:localhost"}},
                },
                "event_id": "$msg3:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567893,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            side_effect=[
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Second message",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$msg1:localhost"}},
                        },
                        "event_id": "$msg2:localhost",
                        "sender": "@mindroom_general:localhost",
                        "origin_server_ts": 1234567892,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "First message",
                            "msgtype": "m.text",
                        },
                        "event_id": "$msg1:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567891,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            ],
        )

        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$msg1:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == ["$msg1:localhost", "$msg2:localhost"]
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_long_reply_chain_keeps_true_root(self, bot: AgentBot) -> None:
        """Long reply chains should keep a stable root instead of drifting."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        chain_length = 55  # Intentionally exceeds the old fixed depth cap of 50.
        newest_parent_id = f"$msg{chain_length}:localhost"
        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": f"{chain_length + 1}th message in reply-only chain",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": newest_parent_id}},
                },
                "event_id": f"$msg{chain_length + 1}:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890 + chain_length + 1,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        responses: list[nio.RoomGetEventResponse] = []
        for i in range(chain_length, 0, -1):
            content = {"body": f"Message {i}", "msgtype": "m.text"}
            if i > 1:
                content["m.relates_to"] = {"m.in_reply_to": {"event_id": f"$msg{i - 1}:localhost"}}

            responses.append(
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": content,
                        "event_id": f"$msg{i}:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567880 + i,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            )

        bot.client.room_get_event = AsyncMock(side_effect=responses)

        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            context = await bot._extract_message_context(room, event)
            # Re-resolving should use cached reply-chain nodes and roots.
            context_cached = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context_cached.is_thread is True
        assert context.thread_id == "$msg1:localhost"
        assert context_cached.thread_id == "$msg1:localhost"
        assert len(context.thread_history) == chain_length
        assert context.thread_history[0]["event_id"] == "$msg1:localhost"
        assert context.thread_history[-1]["event_id"] == newest_parent_id
        assert bot.client.room_get_event.await_count == chain_length
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_reply_chain_cycle_stops_cleanly(self, bot: AgentBot) -> None:
        """Cycle traversal should terminate without looping forever."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Cycle edge case",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$msg3:localhost"}},
                },
                "event_id": "$incoming:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567896,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            side_effect=[
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Message 3",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$msg2:localhost"}},
                        },
                        "event_id": "$msg3:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567895,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Message 2",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$msg3:localhost"}},
                        },
                        "event_id": "$msg2:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567894,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            ],
        )

        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$msg2:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == ["$msg2:localhost", "$msg3:localhost"]
        assert bot.client.room_get_event.await_count == 2
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_reply_chain_traversal_limit_warns_and_falls_back(self, bot: AgentBot) -> None:
        """Traversal limit should warn and return the oldest resolved event as fallback root."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Deep non-cyclic reply chain",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$msg6:localhost"}},
                },
                "event_id": "$incoming:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567897,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        responses: list[nio.RoomGetEventResponse] = []
        for i in range(6, 0, -1):
            content = {"body": f"Message {i}", "msgtype": "m.text"}
            if i > 1:
                content["m.relates_to"] = {"m.in_reply_to": {"event_id": f"$msg{i - 1}:localhost"}}
            responses.append(
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": content,
                        "event_id": f"$msg{i}:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567880 + i,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            )

        bot.client.room_get_event = AsyncMock(side_effect=responses)

        bot._reply_chain.traversal_limit = 3
        with (
            patch.object(bot.logger, "warning") as mock_warning,
            patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch,
        ):
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$msg4:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == [
            "$msg4:localhost",
            "$msg5:localhost",
            "$msg6:localhost",
        ]
        assert bot.client.room_get_event.await_count == 3
        mock_warning.assert_called_once()
        assert mock_warning.call_args.kwargs["traversal_limit"] == 3
        assert mock_warning.call_args.kwargs["traversed_events"] == 3
        assert mock_warning.call_args.kwargs["last_event_id"] == "$msg4:localhost"
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_deeper_traversal_overrides_stale_limit_cache(self, bot: AgentBot) -> None:
        """A successful deeper traversal must override a stale root cached from a prior limit hit."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        def _make_chain_responses(length: int) -> list[nio.RoomGetEventResponse]:
            responses: list[nio.RoomGetEventResponse] = []
            for i in range(length, 0, -1):
                content: dict[str, Any] = {"body": f"Message {i}", "msgtype": "m.text"}
                if i > 1:
                    content["m.relates_to"] = {"m.in_reply_to": {"event_id": f"$msg{i - 1}:localhost"}}
                responses.append(
                    nio.RoomGetEventResponse.from_dict(
                        {
                            "content": content,
                            "event_id": f"$msg{i}:localhost",
                            "sender": "@user:localhost",
                            "origin_server_ts": 1234567880 + i,
                            "room_id": "!test:localhost",
                            "type": "m.room.message",
                        },
                    ),
                )
            return responses

        # --- First resolve: limit=3 starting from $msg6, caches stale root $msg4 ---
        event1 = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "First incoming",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$msg6:localhost"}},
                },
                "event_id": "$incoming1:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567900,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(side_effect=_make_chain_responses(6))
        bot._reply_chain.traversal_limit = 3
        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            ctx1 = await bot._extract_message_context(room, event1)

        assert ctx1.thread_id == "$msg4:localhost"  # stale root from limit hit
        mock_fetch.assert_not_called()

        # --- Second resolve: default limit, new event replies to $msg6 (overlapping cached events) ---
        bot._reply_chain.traversal_limit = 500
        event2 = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Second incoming",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$msg6:localhost"}},
                },
                "event_id": "$incoming2:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567901,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        # Node cache already has $msg6..$msg4; supply $msg3..$msg1 for the deeper walk
        bot.client.room_get_event = AsyncMock(side_effect=_make_chain_responses(3))
        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            ctx2 = await bot._extract_message_context(room, event2)

        assert ctx2.is_thread is True
        assert ctx2.thread_id == "$msg1:localhost"  # true root, not stale $msg4
        assert ctx2.thread_history[0]["event_id"] == "$msg1:localhost"
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_reply_chain_caches_stay_bounded(self, bot: AgentBot) -> None:
        """Reply-chain caches should stay bounded by LRU eviction."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        chain_length = 12
        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Check bounded caches",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": f"$msg{chain_length}:localhost"}},
                },
                "event_id": "$incoming:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567990,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        responses: list[nio.RoomGetEventResponse] = []
        for i in range(chain_length, 0, -1):
            content = {"body": f"Message {i}", "msgtype": "m.text"}
            if i > 1:
                content["m.relates_to"] = {"m.in_reply_to": {"event_id": f"$msg{i - 1}:localhost"}}
            responses.append(
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": content,
                        "event_id": f"$msg{i}:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567900 + i,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            )

        bot.client.room_get_event = AsyncMock(side_effect=responses)

        bot._reply_chain.nodes.maxsize = 5
        bot._reply_chain.roots.maxsize = 5
        with patch("mindroom.bot.fetch_thread_history", AsyncMock()) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$msg1:localhost"
        assert len(bot._reply_chain.nodes) <= 5
        assert len(bot._reply_chain.roots) <= 5
        mock_fetch.assert_not_called()

    @pytest.mark.asyncio
    async def test_extract_context_preserves_plain_replies_before_thread_link(self, bot: AgentBot) -> None:
        """Reply-chain messages should be preserved when chain eventually points to a thread."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Newest plain reply from non-thread client",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$plain2:localhost"}},
                },
                "event_id": "$plain3:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567895,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            side_effect=[
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Second plain reply",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$plain1:localhost"}},
                        },
                        "event_id": "$plain2:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567894,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "First plain reply",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$thread_msg:localhost"}},
                        },
                        "event_id": "$plain1:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567893,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Earlier threaded message",
                            "msgtype": "m.text",
                            "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                        },
                        "event_id": "$thread_msg:localhost",
                        "sender": "@mindroom_general:localhost",
                        "origin_server_ts": 1234567892,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            ],
        )

        thread_history = [
            {"event_id": "$thread_root:localhost", "body": "Thread root"},
            {"event_id": "$thread_msg:localhost", "body": "Earlier threaded message"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=thread_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$thread_root:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == [
            "$thread_root:localhost",
            "$thread_msg:localhost",
            "$plain1:localhost",
            "$plain2:localhost",
        ]
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$thread_root:localhost")

    @pytest.mark.asyncio
    async def test_extract_context_preserves_plain_replies_across_thread_reentries(self, bot: AgentBot) -> None:
        """Plain replies should remain in context even when chain re-enters threaded events."""
        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Newest plain reply from non-thread client",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$p2:localhost"}},
                },
                "event_id": "$incoming:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567896,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            side_effect=[
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Second plain reply",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$t2:localhost"}},
                        },
                        "event_id": "$p2:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567895,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "Thread reply after plain interleave",
                            "msgtype": "m.text",
                            "m.relates_to": {
                                "rel_type": "m.thread",
                                "event_id": "$root:localhost",
                                "m.in_reply_to": {"event_id": "$p1:localhost"},
                            },
                        },
                        "event_id": "$t2:localhost",
                        "sender": "@mindroom_general:localhost",
                        "origin_server_ts": 1234567894,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "First plain interleaved reply",
                            "msgtype": "m.text",
                            "m.relates_to": {"m.in_reply_to": {"event_id": "$t1:localhost"}},
                        },
                        "event_id": "$p1:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567893,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {
                            "body": "First threaded reply",
                            "msgtype": "m.text",
                            "m.relates_to": {
                                "rel_type": "m.thread",
                                "event_id": "$root:localhost",
                                "m.in_reply_to": {"event_id": "$root:localhost"},
                            },
                        },
                        "event_id": "$t1:localhost",
                        "sender": "@mindroom_general:localhost",
                        "origin_server_ts": 1234567892,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
                nio.RoomGetEventResponse.from_dict(
                    {
                        "content": {"body": "Thread root", "msgtype": "m.text"},
                        "event_id": "$root:localhost",
                        "sender": "@user:localhost",
                        "origin_server_ts": 1234567891,
                        "room_id": "!test:localhost",
                        "type": "m.room.message",
                    },
                ),
            ],
        )

        thread_history = [
            {"event_id": "$root:localhost", "body": "Thread root"},
            {"event_id": "$t1:localhost", "body": "First threaded reply"},
            {"event_id": "$t2:localhost", "body": "Thread reply after plain interleave"},
        ]
        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=thread_history)) as mock_fetch:
            context = await bot._extract_message_context(room, event)

        assert context.is_thread is True
        assert context.thread_id == "$root:localhost"
        assert [msg["event_id"] for msg in context.thread_history] == [
            "$root:localhost",
            "$t1:localhost",
            "$p1:localhost",
            "$t2:localhost",
            "$p2:localhost",
        ]
        mock_fetch.assert_awaited_once_with(bot.client, room.room_id, "$root:localhost")

    def test_merge_thread_and_chain_history_preserves_chronological_order(self) -> None:
        """Merged context should preserve chronological order for interleaved plain replies."""
        thread_history = [
            {"event_id": "$root:localhost", "body": "Thread root"},
            {"event_id": "$t1:localhost", "body": "First threaded reply"},
            {"event_id": "$t2:localhost", "body": "Thread reply after plain interleave"},
        ]
        chain_history = [
            {"event_id": "$root:localhost", "body": "Thread root"},
            {"event_id": "$t1:localhost", "body": "First threaded reply"},
            {"event_id": "$p1:localhost", "body": "First plain interleaved reply"},
            {"event_id": "$t2:localhost", "body": "Thread reply after plain interleave"},
            {"event_id": "$p2:localhost", "body": "Second plain reply"},
        ]

        merged = _merge_thread_and_chain_history(thread_history, chain_history)

        assert [msg["event_id"] for msg in merged] == [
            "$root:localhost",
            "$t1:localhost",
            "$p1:localhost",
            "$t2:localhost",
            "$p2:localhost",
        ]

    @pytest.mark.asyncio
    async def test_command_as_reply_doesnt_cause_thread_error(self, tmp_path: Path) -> None:
        """Test that commands sent as replies don't cause threading errors."""
        # Create a router bot to handle commands
        agent_user = AgentMatrixUser(
            user_id="@mindroom_router:localhost",
            password=TEST_PASSWORD,
            display_name="Router",
            agent_name="router",
        )

        config = Config(
            agents={"router": AgentConfig(display_name="Router", rooms=["!test:localhost"])},
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="ollama", id="test-model")},
            router=RouterConfig(model="default"),
        )

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            rooms=["!test:localhost"],
            enable_streaming=False,
            config=config,
        )

        # Mock the orchestrator
        mock_orchestrator = MagicMock()
        mock_orchestrator.current_config = config
        bot.orchestrator = mock_orchestrator

        # Create a mock client
        bot.client = AsyncMock(spec=nio.AsyncClient)
        bot.client.rooms = {}
        bot.client.user_id = "@mindroom_router:localhost"
        bot.client.homeserver = "http://localhost:8008"

        # Initialize components that depend on client
        bot.response_tracker = MagicMock()
        bot.response_tracker.has_responded.return_value = False

        # Mock the agent to return a response
        mock_agent = MagicMock()
        mock_response = MagicMock()
        mock_response.content = "I can help you with that!"

        # Make the agent's arun method return the response
        async def mock_arun(*_args: object, **_kwargs: object) -> MagicMock:
            return mock_response

        mock_agent.arun = mock_arun

        with patch("mindroom.bot.create_agent", return_value=mock_agent):
            room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
            room.name = "Test Room"

            # Create a command that's a reply to another message (not in a thread)
            event = nio.RoomMessageText.from_dict(
                {
                    "content": {
                        "body": "!help",
                        "msgtype": "m.text",
                        "m.relates_to": {"m.in_reply_to": {"event_id": "$some_other_msg:localhost"}},
                    },
                    "event_id": "$cmd_reply:localhost",
                    "sender": "@user:localhost",
                    "origin_server_ts": 1234567890,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            )

            # Mock the bot's response - it should succeed
            bot.client.room_send = AsyncMock(
                return_value=nio.RoomSendResponse.from_dict(
                    {"event_id": "$response:localhost"},
                    room_id="!test:localhost",
                ),
            )

            # Process the command
            await bot._on_message(room, event)

            # The bot should send an error message about needing threads
            bot.client.room_send.assert_called_once()

            # Check the content
            call_args = bot.client.room_send.call_args
            content = call_args.kwargs["content"]

            # The error response should create a thread from the message the command is replying to
            # Since the command is a reply to $some_other_msg:localhost, that becomes the thread root
            assert "m.relates_to" in content
            assert content["m.relates_to"]["rel_type"] == "m.thread"
            # Thread root should be the message the command was replying to
            assert content["m.relates_to"]["event_id"] == "$some_other_msg:localhost"
            # Should reply to the command message
            assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$cmd_reply:localhost"

    @pytest.mark.asyncio
    async def test_command_in_thread_works_correctly(self, tmp_path: Path) -> None:
        """Test that commands in threads work without errors."""
        # Create a router bot to handle commands
        agent_user = AgentMatrixUser(
            user_id="@mindroom_router:localhost",
            password=TEST_PASSWORD,
            display_name="Router",
            agent_name="router",
        )

        config = Config(
            agents={"router": AgentConfig(display_name="Router", rooms=["!test:localhost"])},
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="ollama", id="test-model")},
            router=RouterConfig(model="default"),
        )

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            rooms=["!test:localhost"],
            enable_streaming=False,
            config=config,
        )
        # Mock the orchestrator
        mock_orchestrator = MagicMock()
        mock_orchestrator.current_config = config
        bot.orchestrator = mock_orchestrator

        # Create a mock client
        bot.client = AsyncMock(spec=nio.AsyncClient)
        bot.client.rooms = {}
        bot.client.user_id = "@mindroom_router:localhost"
        bot.client.homeserver = "http://localhost:8008"

        # Initialize components that depend on client
        bot.response_tracker = MagicMock()
        bot.response_tracker.has_responded.return_value = False

        # Mock the agent to return a response
        mock_agent = MagicMock()
        mock_response = MagicMock()
        mock_response.content = "I can help you with that!"

        # Make the agent's arun method return the response
        async def mock_arun(*_args: object, **_kwargs: object) -> MagicMock:
            return mock_response

        mock_agent.arun = mock_arun

        with patch("mindroom.bot.create_agent", return_value=mock_agent):
            room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
            room.name = "Test Room"

            # Create a command in a thread
            event = nio.RoomMessageText.from_dict(
                {
                    "content": {
                        "body": "!list_schedules",
                        "msgtype": "m.text",
                        "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                    },
                    "event_id": "$cmd_thread:localhost",
                    "sender": "@user:localhost",
                    "origin_server_ts": 1234567890,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            )

            # Mock room_get_state for list_schedules command
            bot.client.room_get_state = AsyncMock(
                return_value=nio.RoomGetStateResponse.from_dict(
                    [],  # No scheduled tasks
                    room_id="!test:localhost",
                ),
            )

            # Mock the bot's response
            bot.client.room_send = AsyncMock(
                return_value=nio.RoomSendResponse.from_dict(
                    {"event_id": "$response:localhost"},
                    room_id="!test:localhost",
                ),
            )

            # Process the command
            await bot._on_message(room, event)

            # The bot should respond
            bot.client.room_send.assert_called_once()

            # Check the content
            call_args = bot.client.room_send.call_args
            content = call_args.kwargs["content"]

            # The response should be in the same thread
            assert "m.relates_to" in content
            assert content["m.relates_to"]["rel_type"] == "m.thread"
            assert content["m.relates_to"]["event_id"] == "$thread_root:localhost"
            assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$cmd_thread:localhost"

    @pytest.mark.asyncio
    async def test_command_reply_to_thread_message_uses_existing_thread_root(self, tmp_path: Path) -> None:
        """Plain replies to a threaded message should keep command replies in that thread."""
        agent_user = AgentMatrixUser(
            user_id="@mindroom_router:localhost",
            password=TEST_PASSWORD,
            display_name="Router",
            agent_name="router",
        )

        config = Config(
            agents={"router": AgentConfig(display_name="Router", rooms=["!test:localhost"])},
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="ollama", id="test-model")},
            router=RouterConfig(model="default"),
        )

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            rooms=["!test:localhost"],
            enable_streaming=False,
            config=config,
        )
        mock_orchestrator = MagicMock()
        mock_orchestrator.current_config = config
        bot.orchestrator = mock_orchestrator

        bot.client = AsyncMock(spec=nio.AsyncClient)
        bot.client.rooms = {}
        bot.client.user_id = "@mindroom_router:localhost"
        bot.client.homeserver = "http://localhost:8008"

        bot.response_tracker = MagicMock()
        bot.response_tracker.has_responded.return_value = False

        room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "!help",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$thread_msg:localhost"}},
                },
                "event_id": "$cmd_reply_plain:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            return_value=nio.RoomGetEventResponse.from_dict(
                {
                    "content": {
                        "body": "Agent thread message",
                        "msgtype": "m.text",
                        "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                    },
                    "event_id": "$thread_msg:localhost",
                    "sender": "@mindroom_general:localhost",
                    "origin_server_ts": 1234567890,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            ),
        )

        bot.client.room_send = AsyncMock(
            return_value=nio.RoomSendResponse.from_dict(
                {"event_id": "$response:localhost"},
                room_id="!test:localhost",
            ),
        )

        with patch("mindroom.bot.fetch_thread_history", AsyncMock(return_value=[])):
            await bot._on_message(room, event)

        bot.client.room_send.assert_called_once()
        content = bot.client.room_send.call_args.kwargs["content"]
        assert content["m.relates_to"]["rel_type"] == "m.thread"
        assert content["m.relates_to"]["event_id"] == "$thread_root:localhost"
        assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$cmd_reply_plain:localhost"

    @pytest.mark.asyncio
    async def test_router_routing_reply_to_thread_message_uses_existing_thread_root(self, tmp_path: Path) -> None:
        """Router routing should resolve plain replies back to the real thread root."""
        agent_user = AgentMatrixUser(
            user_id="@mindroom_router:localhost",
            password=TEST_PASSWORD,
            display_name="Router",
            agent_name="router",
        )

        config = Config(
            agents={
                "general": AgentConfig(display_name="General", rooms=["!test:localhost"]),
            },
            teams={},
            room_models={},
            models={"default": ModelConfig(provider="ollama", id="test-model")},
            router=RouterConfig(model="default"),
        )

        bot = AgentBot(
            agent_user=agent_user,
            storage_path=tmp_path,
            rooms=["!test:localhost"],
            enable_streaming=False,
            config=config,
        )
        mock_orchestrator = MagicMock()
        mock_orchestrator.current_config = config
        bot.orchestrator = mock_orchestrator

        bot.client = AsyncMock(spec=nio.AsyncClient)
        bot.client.rooms = {}
        bot.client.user_id = "@mindroom_router:localhost"
        bot.client.homeserver = "http://localhost:8008"

        bot.response_tracker = MagicMock()
        bot.response_tracker.has_responded.return_value = False

        room = MagicMock(spec=nio.MatrixRoom)
        room.room_id = "!test:localhost"
        room.name = "Test Room"

        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "Can someone help with this?",
                    "msgtype": "m.text",
                    "m.relates_to": {"m.in_reply_to": {"event_id": "$thread_msg:localhost"}},
                },
                "event_id": "$plain_reply:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        bot.client.room_get_event = AsyncMock(
            return_value=nio.RoomGetEventResponse.from_dict(
                {
                    "content": {
                        "body": "Earlier message in thread",
                        "msgtype": "m.text",
                        "m.relates_to": {"rel_type": "m.thread", "event_id": "$thread_root:localhost"},
                    },
                    "event_id": "$thread_msg:localhost",
                    "sender": "@mindroom_general:localhost",
                    "origin_server_ts": 1234567889,
                    "room_id": "!test:localhost",
                    "type": "m.room.message",
                },
            ),
        )

        with (
            patch("mindroom.bot.suggest_agent_for_message", AsyncMock(return_value="general")),
            patch("mindroom.bot.get_latest_thread_event_id_if_needed", AsyncMock(return_value="$latest:localhost")),
            patch("mindroom.bot.send_message", AsyncMock(return_value="$router_response:localhost")) as mock_send,
        ):
            await bot._handle_ai_routing(room, event, thread_history=[], thread_id="$thread_root:localhost")

        mock_send.assert_awaited_once()
        bot.client.room_get_event.assert_not_called()
        content = mock_send.call_args.args[2]
        assert content["m.relates_to"]["rel_type"] == "m.thread"
        assert content["m.relates_to"]["event_id"] == "$thread_root:localhost"
        assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$plain_reply:localhost"

    @pytest.mark.asyncio
    async def test_message_with_multiple_relations_handled_correctly(self, bot: AgentBot) -> None:
        """Test that messages with complex relations are handled properly."""
        room = nio.MatrixRoom(room_id="!test:localhost", own_user_id=bot.client.user_id)
        room.name = "Test Room"

        # Create a message that's both in a thread AND a reply (complex relations)
        event = nio.RoomMessageText.from_dict(
            {
                "content": {
                    "body": "@mindroom_general Complex question?",
                    "msgtype": "m.text",
                    "m.mentions": {"user_ids": ["@mindroom_general:localhost"]},
                    "m.relates_to": {
                        "rel_type": "m.thread",
                        "event_id": "$thread_root:localhost",
                        "m.in_reply_to": {"event_id": "$previous_msg:localhost"},
                    },
                },
                "event_id": "$complex_msg:localhost",
                "sender": "@user:localhost",
                "origin_server_ts": 1234567890,
                "room_id": "!test:localhost",
                "type": "m.room.message",
            },
        )

        # Mock the bot's response
        bot.client.room_send = AsyncMock(
            return_value=nio.RoomSendResponse.from_dict({"event_id": "$response:localhost"}, room_id="!test:localhost"),
        )

        # Mock thread history
        bot.client.room_messages = AsyncMock(
            return_value=nio.RoomMessagesResponse.from_dict(
                {"chunk": [], "start": "s1", "end": "e1"},
                room_id="!test:localhost",
            ),
        )

        # Initialize response tracking
        bot.response_tracker.has_responded.return_value = False

        # Mock interactive.handle_text_response and generate_response
        with (
            patch("mindroom.bot.interactive.handle_text_response", AsyncMock(return_value=None)),
            patch.object(bot, "_generate_response") as mock_generate,
        ):
            # Process the message
            await bot._on_message(room, event)

            # Check that _generate_response was called
            mock_generate.assert_called_once()

            # Now simulate the response being sent
            await bot._send_response(
                room.room_id,
                event.event_id,
                "I can help with that complex question!",
                "$thread_root:localhost",
            )

        # Verify the bot sent a response
        bot.client.room_send.assert_called_once()

        # Check the content
        call_args = bot.client.room_send.call_args
        content = call_args.kwargs["content"]

        # The response should maintain the thread context
        assert "m.relates_to" in content
        assert content["m.relates_to"]["rel_type"] == "m.thread"
        assert content["m.relates_to"]["event_id"] == "$thread_root:localhost"
        assert content["m.relates_to"]["m.in_reply_to"]["event_id"] == "$complex_msg:localhost"



"""Provisioning helpers for hosted local-MindRoom registration flows."""

from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Literal

import httpx

from mindroom.constants import MATRIX_SSL_VERIFY
from mindroom.matrix.client import matrix_startup_error


def provisioning_url_from_env() -> str | None:
    """Get hosted provisioning API base URL from environment if configured."""
    url = os.getenv("MINDROOM_PROVISIONING_URL", "").strip()
    return url.rstrip("/") or None


def registration_token_from_env() -> str | None:
    """Get MATRIX_REGISTRATION_TOKEN from environment if configured."""
    token = os.getenv("MATRIX_REGISTRATION_TOKEN", "").strip()
    return token or None


def _local_provisioning_client_credentials_from_env() -> tuple[str, str] | None:
    """Get local provisioning client credentials from environment if configured."""
    client_id = os.getenv("MINDROOM_LOCAL_CLIENT_ID", "").strip()
    client_secret = os.getenv("MINDROOM_LOCAL_CLIENT_SECRET", "").strip()
    if not client_id and not client_secret:
        return None
    if not client_id or not client_secret:
        msg = (
            "Provisioning credentials are incomplete. "
            "Set both MINDROOM_LOCAL_CLIENT_ID and MINDROOM_LOCAL_CLIENT_SECRET, "
            "or run `mindroom connect --pair-code ...` again."
        )
        raise matrix_startup_error(msg, permanent=True)
    return client_id, client_secret


def required_local_provisioning_client_credentials_for_registration(
    *,
    provisioning_url: str | None,
    registration_token: str | None,
) -> tuple[str, str] | None:
    """Resolve required local provisioning credentials when using hosted registration."""
    if registration_token or not provisioning_url:
        return None

    creds = _local_provisioning_client_credentials_from_env()
    if creds is None:
        msg = (
            "MINDROOM_PROVISIONING_URL is set but local client credentials are missing. "
            "Run `mindroom connect --pair-code ...` first."
        )
        raise matrix_startup_error(msg, permanent=True)
    return creds


@dataclass(frozen=True)
class _ProvisioningRegisterResult:
    """Result returned by the provisioning register-agent endpoint."""

    status: Literal["created", "user_in_use"]
    user_id: str


async def register_user_via_provisioning_service(
    *,
    provisioning_url: str,
    client_id: str,
    client_secret: str,
    homeserver: str,
    username: str,
    password: str,
    display_name: str,
) -> _ProvisioningRegisterResult:
    """Register an agent account via provisioning service server-side flow."""
    url = f"{provisioning_url}/v1/local-mindroom/register-agent"
    headers = {
        "X-Local-MindRoom-Client-Id": client_id,
        "X-Local-MindRoom-Client-Secret": client_secret,
    }
    payload = {
        "homeserver": homeserver.rstrip("/"),
        "username": username,
        "password": password,
        "display_name": display_name,
    }
    try:
        async with httpx.AsyncClient(timeout=10, verify=MATRIX_SSL_VERIFY) as client:
            response = await client.post(url, json=payload, headers=headers)
    except httpx.HTTPError as exc:
        msg = f"Could not reach provisioning service ({provisioning_url}): {exc}"
        raise ValueError(msg) from exc

    if not response.is_success:
        detail = response.text.strip() or "unknown error"
        if response.status_code in {401, 403}:
            msg = "Provisioning credentials are invalid or revoked. Run `mindroom connect --pair-code ...` again."
            raise matrix_startup_error(msg, permanent=True)
        if response.status_code == 404:
            msg = (
                "Provisioning service does not support /register-agent yet. "
                "Deploy the latest local provisioning service."
            )
            raise matrix_startup_error(msg, permanent=True)
        msg = f"Provisioning service returned HTTP {response.status_code}: {detail}"
        raise ValueError(msg)

    try:
        body = response.json()
    except ValueError as exc:
        msg = "Provisioning service returned invalid JSON while registering agent."
        raise matrix_startup_error(msg, permanent=True) from exc

    if not isinstance(body, dict):
        msg = "Provisioning service returned invalid register-agent payload."
        raise matrix_startup_error(msg, permanent=True)

    status = body.get("status")
    user_id = body.get("user_id")
    if status not in {"created", "user_in_use"}:
        msg = "Provisioning service response missing valid status for register-agent."
        raise matrix_startup_error(msg, permanent=True)
    if not isinstance(user_id, str) or not user_id.strip():
        msg = "Provisioning service response missing user_id for register-agent."
        raise matrix_startup_error(msg, permanent=True)

    return _ProvisioningRegisterResult(status=status, user_id=user_id.strip())



"""Resource and manifest helpers for the Kubernetes worker backend."""

from __future__ import annotations

import hashlib
import importlib
import os
import time
from typing import TYPE_CHECKING, Protocol, cast

from mindroom.credentials import SHARED_CREDENTIALS_PATH_ENV
from mindroom.workers.backend import WorkerBackendError

if TYPE_CHECKING:
    from collections.abc import Callable

    from mindroom.workers.models import WorkerStatus

    from .kubernetes_config import _KubernetesWorkerBackendConfig

_DEFAULT_NAME_PREFIX = "mindroom-worker"
_DEFAULT_RESOURCE_REQUESTS = {"memory": "256Mi", "cpu": "100m"}
_DEFAULT_RESOURCE_LIMITS = {"memory": "1Gi", "cpu": "500m"}
_READY_POLL_INTERVAL_SECONDS = 1.0
_HOSTNAME_ENV = "HOSTNAME"

ANNOTATION_CREATED_AT = "mindroom.ai/created-at"
ANNOTATION_LAST_USED_AT = "mindroom.ai/last-used-at"
ANNOTATION_LAST_STARTED_AT = "mindroom.ai/last-started-at"
ANNOTATION_STARTUP_COUNT = "mindroom.ai/startup-count"
ANNOTATION_FAILURE_COUNT = "mindroom.ai/failure-count"
ANNOTATION_FAILURE_REASON = "mindroom.ai/failure-reason"
ANNOTATION_WORKER_KEY = "mindroom.ai/worker-key"
ANNOTATION_WORKER_STATUS = "mindroom.ai/worker-status"
ANNOTATION_STATE_SUBPATH = "mindroom.ai/state-subpath"

_LABEL_COMPONENT = "mindroom.ai/component"
_LABEL_COMPONENT_VALUE = "worker"
_LABEL_MANAGED_BY = "app.kubernetes.io/managed-by"
_LABEL_MANAGED_BY_VALUE = "mindroom"
_LABEL_NAME = "app.kubernetes.io/name"
_LABEL_NAME_VALUE = "mindroom-worker"
_LABEL_WORKER_ID = "mindroom.ai/worker-id"

_CONTAINER_NAME = "sandbox-runner"
_TOKEN_ENV_NAME = "MINDROOM_SANDBOX_PROXY_TOKEN"  # noqa: S105
_RUNNER_PORT_ENV_NAME = "MINDROOM_SANDBOX_RUNNER_PORT"
_DEDICATED_WORKER_KEY_ENV = "MINDROOM_SANDBOX_DEDICATED_WORKER_KEY"
_DEDICATED_WORKER_ROOT_ENV = "MINDROOM_SANDBOX_DEDICATED_WORKER_ROOT"


class _ApiStatusError(Exception):
    status: int


class _KubernetesMetadata(Protocol):
    name: str
    annotations: dict[str, str] | None
    labels: dict[str, str]
    generation: int | None
    uid: str | None


class _KubernetesDeploymentSpec(Protocol):
    replicas: int | None


class _KubernetesDeploymentStatus(Protocol):
    ready_replicas: int | None
    observed_generation: int | None


class KubernetesDeployment(Protocol):
    """Minimal Deployment surface used by the backend."""

    metadata: _KubernetesMetadata
    spec: _KubernetesDeploymentSpec
    status: _KubernetesDeploymentStatus


class _KubernetesPodSpec(Protocol):
    node_name: str | None


class _KubernetesPod(Protocol):
    spec: _KubernetesPodSpec


class _KubernetesDeploymentList(Protocol):
    items: list[KubernetesDeployment] | None


class _AppsApiProtocol(Protocol):
    def read_namespaced_deployment(self, name: str, namespace: str) -> KubernetesDeployment: ...

    def create_namespaced_deployment(self, namespace: str, body: dict[str, object]) -> KubernetesDeployment: ...

    def patch_namespaced_deployment(
        self,
        name: str,
        namespace: str,
        body: dict[str, object],
    ) -> KubernetesDeployment: ...

    def delete_namespaced_deployment(self, name: str, namespace: str) -> None: ...

    def list_namespaced_deployment(self, namespace: str, label_selector: str) -> _KubernetesDeploymentList: ...


class _CoreApiProtocol(Protocol):
    def read_namespaced_service(self, name: str, namespace: str) -> object: ...

    def create_namespaced_service(self, namespace: str, body: dict[str, object]) -> object: ...

    def patch_namespaced_service(self, name: str, namespace: str, body: dict[str, object]) -> object: ...

    def delete_namespaced_service(self, name: str, namespace: str) -> None: ...

    def read_namespaced_pod(self, name: str, namespace: str) -> _KubernetesPod: ...


def worker_id_for_key(worker_key: str, *, prefix: str) -> str:
    """Return a DNS-safe Kubernetes resource name for one worker key."""
    digest = hashlib.sha256(worker_key.encode("utf-8")).hexdigest()[:24]
    normalized_prefix = prefix.strip().lower().strip("-") or _DEFAULT_NAME_PREFIX
    max_prefix_length = 63 - len(digest) - 1
    safe_prefix = normalized_prefix[:max_prefix_length].rstrip("-")
    if not safe_prefix:
        safe_prefix = _DEFAULT_NAME_PREFIX[:max_prefix_length].rstrip("-") or "worker"
    return f"{safe_prefix}-{digest}"


def service_host(service_name: str, namespace: str, port: int) -> str:
    """Return the cluster-local HTTP root for one worker Service."""
    return f"http://{service_name}.{namespace}.svc.cluster.local:{port}"


def parse_annotation_float(annotations: dict[str, str], key: str, default: float) -> float:
    """Parse one float annotation, falling back to a caller-provided default."""
    raw = annotations.get(key)
    if raw is None:
        return default
    try:
        return float(raw)
    except ValueError:
        return default


def parse_annotation_int(annotations: dict[str, str], key: str, default: int = 0) -> int:
    """Parse one integer annotation, falling back to a caller-provided default."""
    raw = annotations.get(key)
    if raw is None:
        return default
    try:
        return int(raw)
    except ValueError:
        return default


def metadata_annotations(
    *,
    worker_key: str,
    state_subpath: str,
    created_at: float,
    last_used_at: float,
    last_started_at: float | None,
    startup_count: int,
    failure_count: int,
    failure_reason: str | None,
    status: WorkerStatus,
) -> dict[str, str]:
    """Build persisted worker lifecycle metadata stored on Deployments."""
    annotations = {
        ANNOTATION_WORKER_KEY: worker_key,
        ANNOTATION_STATE_SUBPATH: state_subpath,
        ANNOTATION_CREATED_AT: str(created_at),
        ANNOTATION_LAST_USED_AT: str(last_used_at),
        ANNOTATION_STARTUP_COUNT: str(startup_count),
        ANNOTATION_FAILURE_COUNT: str(failure_count),
        ANNOTATION_WORKER_STATUS: status,
    }
    if last_started_at is not None:
        annotations[ANNOTATION_LAST_STARTED_AT] = str(last_started_at)
    if failure_reason:
        annotations[ANNOTATION_FAILURE_REASON] = failure_reason
    return annotations


def _labels(*, extra_labels: dict[str, str], worker_id: str) -> dict[str, str]:
    labels = {
        _LABEL_COMPONENT: _LABEL_COMPONENT_VALUE,
        _LABEL_MANAGED_BY: _LABEL_MANAGED_BY_VALUE,
        _LABEL_NAME: _LABEL_NAME_VALUE,
    }
    labels.update(extra_labels)
    labels[_LABEL_WORKER_ID] = worker_id
    return labels


def _list_selector(*, extra_labels: dict[str, str]) -> str:
    selector = {
        _LABEL_COMPONENT: _LABEL_COMPONENT_VALUE,
        _LABEL_MANAGED_BY: _LABEL_MANAGED_BY_VALUE,
        _LABEL_NAME: _LABEL_NAME_VALUE,
    }
    selector.update(extra_labels)
    return ",".join(f"{key}={value}" for key, value in sorted(selector.items()))


class KubernetesResourceManager:
    """Own Kubernetes API access, manifest construction, and cached cluster metadata."""

    def __init__(self, *, config: _KubernetesWorkerBackendConfig, auth_token: str | None) -> None:
        """Initialize one resource manager for a concrete backend configuration."""
        self.config = config
        self.auth_token = auth_token
        self.apps_api: _AppsApiProtocol | None = None
        self.core_api: _CoreApiProtocol | None = None
        self.api_exception_cls: type[_ApiStatusError] | None = None
        self._control_plane_node_name: str | None = None
        self._control_plane_node_name_loaded = False
        self._owner_reference: dict[str, object] | None = None
        self._owner_reference_loaded = False

    @property
    def _apps(self) -> _AppsApiProtocol:
        self._load_clients()
        assert self.apps_api is not None
        return self.apps_api

    @property
    def _core(self) -> _CoreApiProtocol:
        self._load_clients()
        assert self.core_api is not None
        return self.core_api

    @property
    def _api_exception(self) -> type[_ApiStatusError]:
        self._load_clients()
        assert self.api_exception_cls is not None
        return self.api_exception_cls

    def list_deployments(self) -> list[KubernetesDeployment]:
        """List managed worker Deployments in this namespace."""
        response = self._apps.list_namespaced_deployment(
            self.config.namespace,
            label_selector=_list_selector(extra_labels=self.config.extra_labels),
        )
        return list(response.items or [])

    def read_deployment(self, deployment_name: str) -> KubernetesDeployment | None:
        """Read one Deployment, returning ``None`` for 404s."""
        try:
            return self._apps.read_namespaced_deployment(deployment_name, self.config.namespace)
        except self._api_exception as exc:
            if exc.status == 404:
                return None
            raise

    def apply_service(self, worker_id: str) -> None:
        """Create-or-patch one worker Service."""
        self._apply_object(
            read_fn=self._core.read_namespaced_service,
            create_fn=self._core.create_namespaced_service,
            patch_fn=self._core.patch_namespaced_service,
            resource_name=worker_id,
            manifest=self._service_manifest(worker_id),
        )

    def apply_deployment(
        self,
        *,
        worker_key: str,
        worker_id: str,
        state_subpath: str,
        annotations: dict[str, str],
        replicas: int,
    ) -> None:
        """Create-or-patch one worker Deployment."""
        self._apply_object(
            read_fn=self._apps.read_namespaced_deployment,
            create_fn=self._apps.create_namespaced_deployment,
            patch_fn=self._apps.patch_namespaced_deployment,
            resource_name=worker_id,
            manifest=self._deployment_manifest(
                worker_key=worker_key,
                worker_id=worker_id,
                state_subpath=state_subpath,
                annotations=annotations,
                replicas=replicas,
            ),
        )

    def patch_deployment(
        self,
        deployment_name: str,
        *,
        replicas: int | None = None,
        annotations: dict[str, str] | None = None,
    ) -> None:
        """Patch Deployment metadata and/or scale."""
        body: dict[str, object] = {}
        if annotations is not None:
            body["metadata"] = {"annotations": annotations}
        if replicas is not None:
            body["spec"] = {"replicas": replicas}
        self._apps.patch_namespaced_deployment(deployment_name, self.config.namespace, body)

    def delete_deployment(self, deployment_name: str) -> None:
        """Delete one worker Deployment, ignoring 404s."""
        self._delete_object(self._apps.delete_namespaced_deployment, deployment_name)

    def delete_service(self, service_name: str) -> None:
        """Delete one worker Service, ignoring 404s."""
        self._delete_object(self._core.delete_namespaced_service, service_name)

    def wait_for_ready(
        self,
        deployment_name: str,
        *,
        timeout_seconds: float,
        deployment_ready_fn: Callable[[KubernetesDeployment], bool],
    ) -> KubernetesDeployment:
        """Poll a worker Deployment until it becomes ready or times out."""
        deadline = time.time() + timeout_seconds
        while True:
            deployment = self.read_deployment(deployment_name)
            if deployment is None:
                msg = f"Kubernetes worker deployment '{deployment_name}' disappeared during startup."
                raise WorkerBackendError(msg)
            if deployment_ready_fn(deployment):
                return deployment
            if time.time() >= deadline:
                msg = f"Kubernetes worker '{deployment_name}' did not become ready within {timeout_seconds:.0f}s."
                raise WorkerBackendError(msg)
            time.sleep(_READY_POLL_INTERVAL_SECONDS)

    def _apply_object(
        self,
        *,
        read_fn: Callable[[str, str], object],
        create_fn: Callable[[str, dict[str, object]], object],
        patch_fn: Callable[[str, str, dict[str, object]], object],
        resource_name: str,
        manifest: dict[str, object],
    ) -> None:
        try:
            read_fn(resource_name, self.config.namespace)
        except self._api_exception as exc:
            if exc.status != 404:
                raise
            try:
                create_fn(self.config.namespace, manifest)
            except self._api_exception as create_exc:
                if create_exc.status != 409:
                    raise
                patch_fn(resource_name, self.config.namespace, manifest)
            return
        patch_fn(resource_name, self.config.namespace, manifest)

    def _delete_object(self, delete_fn: Callable[[str, str], None], resource_name: str) -> None:
        try:
            delete_fn(resource_name, self.config.namespace)
        except self._api_exception as exc:
            if exc.status != 404:
                raise

    def _load_clients(self) -> None:
        if self.apps_api is not None and self.core_api is not None and self.api_exception_cls is not None:
            return
        try:
            kubernetes_config = importlib.import_module("kubernetes.config")
            kubernetes_client = importlib.import_module("kubernetes.client")
            kubernetes_exceptions = importlib.import_module("kubernetes.client.exceptions")
        except ModuleNotFoundError as exc:
            msg = "The 'kubernetes' package is required for the Kubernetes worker backend."
            raise WorkerBackendError(msg) from exc

        try:
            kubernetes_config.load_incluster_config()
        except Exception:
            kubernetes_config.load_kube_config()

        self.apps_api = cast("_AppsApiProtocol", kubernetes_client.AppsV1Api())
        self.core_api = cast("_CoreApiProtocol", kubernetes_client.CoreV1Api())
        self.api_exception_cls = cast("type[_ApiStatusError]", kubernetes_exceptions.ApiException)

    def _service_manifest(self, worker_id: str) -> dict[str, object]:
        worker_labels = _labels(extra_labels=self.config.extra_labels, worker_id=worker_id)
        metadata: dict[str, object] = {
            "name": worker_id,
            "namespace": self.config.namespace,
            "labels": worker_labels,
        }
        owner_reference = self._owner_reference_or_none()
        if owner_reference is not None:
            metadata["ownerReferences"] = [owner_reference]
        return {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": metadata,
            "spec": {
                "selector": worker_labels,
                "ports": [
                    {
                        "name": "api",
                        "port": self.config.worker_port,
                        "targetPort": self.config.worker_port,
                    },
                ],
            },
        }

    def _deployment_manifest(
        self,
        *,
        worker_key: str,
        worker_id: str,
        state_subpath: str,
        annotations: dict[str, str],
        replicas: int,
    ) -> dict[str, object]:
        worker_labels = _labels(extra_labels=self.config.extra_labels, worker_id=worker_id)
        metadata: dict[str, object] = {
            "name": worker_id,
            "namespace": self.config.namespace,
            "labels": worker_labels,
            "annotations": annotations,
        }
        owner_reference = self._owner_reference_or_none()
        if owner_reference is not None:
            metadata["ownerReferences"] = [owner_reference]

        pod_spec: dict[str, object] = {
            "serviceAccountName": self.config.service_account_name,
            "securityContext": {
                "runAsUser": 1000,
                "runAsGroup": 1000,
                "fsGroup": 1000,
                "runAsNonRoot": True,
                "fsGroupChangePolicy": "OnRootMismatch",
                "seccompProfile": {"type": "RuntimeDefault"},
            },
            "containers": [
                {
                    "name": _CONTAINER_NAME,
                    "image": self.config.image,
                    "imagePullPolicy": self.config.image_pull_policy,
                    "command": ["/app/run-sandbox-runner.sh"],
                    "ports": [{"containerPort": self.config.worker_port, "name": "api"}],
                    "env": self._worker_env(worker_key),
                    "volumeMounts": self._volume_mounts(state_subpath),
                    "readinessProbe": {
                        "httpGet": {"path": "/healthz", "port": "api"},
                        "periodSeconds": 5,
                        "failureThreshold": 6,
                    },
                    "livenessProbe": {
                        "httpGet": {"path": "/healthz", "port": "api"},
                        "periodSeconds": 10,
                        "failureThreshold": 6,
                    },
                    "resources": {
                        "requests": dict(_DEFAULT_RESOURCE_REQUESTS),
                        "limits": dict(_DEFAULT_RESOURCE_LIMITS),
                    },
                    "securityContext": {
                        "allowPrivilegeEscalation": False,
                        "capabilities": {"drop": ["ALL"]},
                    },
                },
            ],
            "volumes": self._volumes(),
        }
        node_name = self._worker_node_name_or_none()
        if node_name is not None:
            pod_spec["nodeName"] = node_name

        return {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": metadata,
            "spec": {
                "replicas": replicas,
                "selector": {"matchLabels": worker_labels},
                "template": {
                    "metadata": {"labels": worker_labels},
                    "spec": pod_spec,
                },
            },
        }

    def _worker_env(self, worker_key: str) -> list[dict[str, object]]:
        env: list[dict[str, object]] = [
            {"name": "MINDROOM_SANDBOX_RUNNER_MODE", "value": "true"},
            {"name": "MINDROOM_SANDBOX_RUNNER_EXECUTION_MODE", "value": "subprocess"},
            {"name": _RUNNER_PORT_ENV_NAME, "value": str(self.config.worker_port)},
            {"name": "MINDROOM_STORAGE_PATH", "value": self.config.storage_mount_path},
            {
                "name": SHARED_CREDENTIALS_PATH_ENV,
                "value": f"{self.config.storage_mount_path}/.shared_credentials",
            },
            {"name": _DEDICATED_WORKER_KEY_ENV, "value": worker_key},
            {"name": _DEDICATED_WORKER_ROOT_ENV, "value": self.config.storage_mount_path},
            {"name": "HOME", "value": self.config.storage_mount_path},
        ]
        if self.config.config_map_name is not None:
            env.append({"name": "MINDROOM_CONFIG_PATH", "value": self.config.config_path})
        if self.config.token_secret_name is not None:
            env.append(
                {
                    "name": _TOKEN_ENV_NAME,
                    "valueFrom": {
                        "secretKeyRef": {
                            "name": self.config.token_secret_name,
                            "key": self.config.token_secret_key,
                        },
                    },
                },
            )
        elif self.auth_token is not None:
            env.append({"name": _TOKEN_ENV_NAME, "value": self.auth_token})
        else:
            msg = "A worker auth token is required for Kubernetes workers."
            raise WorkerBackendError(msg)

        for name, value in sorted(self.config.extra_env.items()):
            env.append({"name": name, "value": value})
        return env

    def _volume_mounts(self, state_subpath: str) -> list[dict[str, object]]:
        mounts: list[dict[str, object]] = [
            {
                "name": "worker-storage",
                "mountPath": self.config.storage_mount_path,
                "subPath": state_subpath,
            },
        ]
        if self.config.config_map_name is not None:
            mounts.append(
                {
                    "name": "worker-config",
                    "mountPath": self.config.config_path,
                    "subPath": self.config.config_key,
                    "readOnly": True,
                },
            )
        return mounts

    def _volumes(self) -> list[dict[str, object]]:
        volumes: list[dict[str, object]] = [
            {
                "name": "worker-storage",
                "persistentVolumeClaim": {"claimName": self.config.storage_pvc_name},
            },
        ]
        if self.config.config_map_name is not None:
            volumes.append(
                {
                    "name": "worker-config",
                    "configMap": {"name": self.config.config_map_name},
                },
            )
        return volumes

    def _worker_node_name_or_none(self) -> str | None:
        if self.config.node_name is not None:
            return self.config.node_name
        if not self.config.colocate_with_control_plane_node:
            return None
        if self._control_plane_node_name_loaded:
            return self._control_plane_node_name

        pod_name = os.getenv(_HOSTNAME_ENV, "").strip()
        if not pod_name:
            self._control_plane_node_name_loaded = True
            return None
        try:
            pod = self._core.read_namespaced_pod(pod_name, self.config.namespace)
        except self._api_exception as exc:
            if exc.status != 404:
                raise
            pod = None
        self._control_plane_node_name = None if pod is None else pod.spec.node_name
        self._control_plane_node_name_loaded = True
        return self._control_plane_node_name

    def _owner_reference_or_none(self) -> dict[str, object] | None:
        if self._owner_reference_loaded:
            return self._owner_reference
        if self.config.owner_deployment_name is None:
            self._owner_reference_loaded = True
            return None

        owner_deployment = self.read_deployment(self.config.owner_deployment_name)
        if owner_deployment is None:
            msg = f"Configured Kubernetes worker owner deployment '{self.config.owner_deployment_name}' was not found."
            raise WorkerBackendError(msg)

        owner_uid = owner_deployment.metadata.uid
        if owner_uid is None or not owner_uid.strip():
            msg = (
                f"Configured Kubernetes worker owner deployment '{self.config.owner_deployment_name}' is missing a UID."
            )
            raise WorkerBackendError(msg)

        self._owner_reference = {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "name": self.config.owner_deployment_name,
            "uid": owner_uid,
            "controller": False,
            "blockOwnerDeletion": False,
        }
        self._owner_reference_loaded = True
        return self._owner_reference
